!MNH_LIC Copyright 1994-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
! Modifications:
!  P. Wautelet 20/05/2019: add name argument to ADDnFIELD_ll + new ADD4DFIELD_ll subroutine
!-----------------------------------------------------------------
!     ####################
      MODULE MODI_GET_HALO
!     ####################
!
INTERFACE
!
   SUBROUTINE GET_HALO2(PSRC,TP_PSRC_HALO2_ll)
     !
     USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
     REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC    ! variable at t
     TYPE(HALO2LIST_ll), POINTER         :: TP_PSRC_HALO2_ll          ! halo2 for SRC
     !
   END SUBROUTINE GET_HALO2
END INTERFACE
!
INTERFACE 
   SUBROUTINE GET_HALO(PSRC,HDIR)
     !
     REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC    ! variable at t
     CHARACTER(len=4), OPTIONAL :: HDIR ! to send only halo on X or Y direction
     !
   END SUBROUTINE GET_HALO
END INTERFACE
! 
#ifdef _OPENACC
INTERFACE            
   SUBROUTINE GET_HALO_D(PSRC,HDIR)
     !
     REAL, DIMENSION(:,:,:), INTENT(INOUT)   :: PSRC    ! variable at t
     !$acc declare present (PSRC)
     CHARACTER(len=4), OPTIONAL :: HDIR ! to send only halo on X or Y direction
     !
   END SUBROUTINE GET_HALO_D
END INTERFACE
#endif
!
INTERFACE
SUBROUTINE DEL_HALO2_ll(TPHALO2LIST)
!
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
TYPE(HALO2LIST_ll), POINTER :: TPHALO2LIST ! list of HALO2_lls
!
END SUBROUTINE DEL_HALO2_ll
!
END INTERFACE
!
END MODULE MODI_GET_HALO         
!
!     ###########################################
      SUBROUTINE GET_HALO2(PSRC,TP_PSRC_HALO2_ll)
!     ###########################################
!
USE MODE_ll
USE MODD_ARGSLIST_ll, ONLY : LIST_ll, HALO2LIST_ll
!
IMPLICIT NONE
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC    ! variable at t
TYPE(HALO2LIST_ll), POINTER         :: TP_PSRC_HALO2_ll          ! halo2 for SRC
!
INTEGER                          :: IIU,IJU,IKU            ! domain sizes
TYPE(LIST_ll)     , POINTER      :: TZ_PSRC_ll               ! halo
INTEGER                          :: IERROR                 ! error return code 
!
IIU = SIZE(PSRC,1)
IJU = SIZE(PSRC,2)
IKU = SIZE(PSRC,3)
!
NULLIFY( TZ_PSRC_ll,TP_PSRC_HALO2_ll)
CALL INIT_HALO2_ll(TP_PSRC_HALO2_ll,1,IIU,IJU,IKU)
!
CALL ADD3DFIELD_ll( TZ_PSRC_ll, PSRC, 'GET_HALO2::PSRC' )
CALL UPDATE_HALO_ll(TZ_PSRC_ll,IERROR)
CALL UPDATE_HALO2_ll(TZ_PSRC_ll,TP_PSRC_HALO2_ll,IERROR)
!
!   clean local halo list
!
CALL CLEANLIST_ll(TZ_PSRC_ll)
!
END SUBROUTINE GET_HALO2
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!     #########################
      SUBROUTINE GET_HALO(PSRC,HDIR)
!     #########################
!
USE MODE_ll
USE MODD_ARGSLIST_ll, ONLY : LIST_ll
!
IMPLICIT NONE
!
REAL, DIMENSION(:,:,:), INTENT(IN)  :: PSRC    ! variable at t
CHARACTER(len=4), OPTIONAL :: HDIR ! to send only halo on X or Y direction
!
TYPE(LIST_ll)     , POINTER      :: TZ_PSRC_ll               ! halo
INTEGER                          :: IERROR                 ! error return code 
!
NULLIFY( TZ_PSRC_ll)
!
CALL ADD3DFIELD_ll( TZ_PSRC_ll, PSRC, 'GET_HALO::PSRC' )
CALL UPDATE_HALO_ll(TZ_PSRC_ll,IERROR, HDIR=HDIR )
CALL CLEANLIST_ll(TZ_PSRC_ll)
!
END SUBROUTINE GET_HALO
!-----------------------------------------------------------------------
#ifdef _OPENACC
MODULE MODD_HALO_D
IMPLICIT NONE 
REAL, SAVE , ALLOCATABLE, DIMENSION(:,:,:)  :: ZNORTH_IN, ZSOUTH_IN, ZWEST_IN, ZEAST_IN
REAL, SAVE , ALLOCATABLE, DIMENSION(:,:,:)  :: ZNORTH_OUT, ZSOUTH_OUT, ZWEST_OUT, ZEAST_OUT
!!$acc declare create (ZNORTH_IN, ZSOUTH_IN, ZWEST_IN, ZEAST_IN)
!!$acc declare create (ZNORTH_OUT, ZSOUTH_OUT, ZWEST_OUT, ZEAST_OUT)

LOGICAL, SAVE                               :: GFIRST_GET_HALO_D = .TRUE.

END MODULE MODD_HALO_D
!-------------------------------------------------------------------------------
!     #########################
      SUBROUTINE GET_HALO_D(PSRC,HDIR)
!     #########################
!
USE MODD_HALO_D
USE MODE_ll
USE MODD_ARGSLIST_ll, ONLY : LIST_ll
USE MODD_PARAMETERS, ONLY : JPHEXT
!
USE MODD_IO,        ONLY : GSMONOPROC
USE MODE_MNH_ZWORK, ONLY : GWEST , GEAST, GSOUTH , GNORTH
!
USE MODD_CONF, ONLY : NHALO
USE MODE_DEVICE
USE MODE_MPPDB
!
IMPLICIT NONE
!
REAL, DIMENSION(:,:,:), INTENT(INOUT) :: PSRC    ! variable at t
!$acc declare present (PSRC)
CHARACTER(len=4), OPTIONAL :: HDIR ! to send only halo on X or Y direction
!
TYPE(LIST_ll)     , POINTER      :: TZ_PSRC_ll               ! halo
INTEGER                          :: IERROR                 ! error return code 
INTEGER, SAVE :: IIB,IJB    ! Begining useful area in x,y,z directions
INTEGER, SAVE :: IIE,IJE    ! End useful area in x,y,z directions

INTEGER,SAVE      :: IIU,IJU,IKU
INTEGER,SAVE      :: IHALO_1
INTEGER,PARAMETER :: IS_WEST=1 , IS_EAST=2, IS_SOUTH=3, IS_NORTH=4

LOGICAL      :: LX , LY

!
!LOGICAL, SAVE                               :: GFIRST_GET_HALO_D = .TRUE.
!

!JUANCHECK3D IF (GSMONOPROC) RETURN
!
#define _PW_NOINTERM
NULLIFY( TZ_PSRC_ll)
!
IF (GFIRST_GET_HALO_D ) THEN 
   CALL GET_INDICE_ll(IIB,IJB,IIE,IJE)
   !
   !
   IIU=size(psrc,1)
   IJU=size(psrc,2)
   IKU=size(psrc,3)
   !
   IHALO_1 = NHALO-1
   !
#ifndef _PW_NOINTERM
   ALLOCATE  ( ZSOUTH_IN ( IIB:IIE   , IJB:IJB+IHALO_1   , IKU ) )
   ALLOCATE  ( ZNORTH_IN ( IIB:IIE   , IJE-IHALO_1:IJE   , IKU ) )
   ALLOCATE  ( ZWEST_IN  ( IIB:IIB+IHALO_1   , IJB:IJE   , IKU ) )
   ALLOCATE  ( ZEAST_IN  ( IIE-IHALO_1:IIE   , IJB:IJE   , IKU ) )
   !$acc enter data create (ZNORTH_IN, ZSOUTH_IN, ZWEST_IN, ZEAST_IN)
   !
   ALLOCATE  ( ZSOUTH_OUT (   IIB:IIE   , 1:IJB-1 , IKU ) )
   ALLOCATE  ( ZNORTH_OUT (   IIB:IIE   , IJE+1:IJU , IKU ) )
   ALLOCATE  ( ZWEST_OUT  ( 1:IIB-1 ,   IJB:IJE   , IKU ) )
   ALLOCATE  ( ZEAST_OUT  ( IIE+1:IIU ,   IJB:IJE   , IKU ) )
   !$acc enter data create (ZNORTH_OUT, ZSOUTH_OUT, ZWEST_OUT, ZEAST_OUT)

   CALL INIT_ON_HOST_AND_DEVICE(ZSOUTH_IN,-1e99,'GET_HALO_D::ZSOUTH_IN')
   CALL INIT_ON_HOST_AND_DEVICE(ZNORTH_IN,-1e99,'GET_HALO_D::ZNORTH_IN')
   CALL INIT_ON_HOST_AND_DEVICE(ZWEST_IN,-1e99,'GET_HALO_D::ZWEST_IN')
   CALL INIT_ON_HOST_AND_DEVICE(ZEAST_IN,-1e99,'GET_HALO_D::ZEAST_IN')

   CALL INIT_ON_HOST_AND_DEVICE(ZSOUTH_OUT,-1e99,'GET_HALO_D::ZSOUTH_OUT')
   CALL INIT_ON_HOST_AND_DEVICE(ZNORTH_OUT,-1e99,'GET_HALO_D::ZNORTH_OUT')
   CALL INIT_ON_HOST_AND_DEVICE(ZWEST_OUT,-1e99,'GET_HALO_D::ZWEST_OUT')
   CALL INIT_ON_HOST_AND_DEVICE(ZEAST_OUT,-1e99,'GET_HALO_D::ZEAST_OUT')
#endif
  
   GFIRST_GET_HALO_D = .FALSE.
END IF 


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

LX = .FALSE.
LY = .FALSE. 

IF (.NOT. PRESENT(HDIR) ) THEN
LX = .TRUE.
LY = .TRUE.
ELSE
LX = ( HDIR == "01_X" .OR. HDIR == "S0_X" )
LY = ( HDIR == "01_Y" .OR. HDIR == "S0_Y" )
!!$print *,"IIB=",IIB," HDIR=",HDIR," LX=",LX," LY=",LY ; call flush(6)
END IF


!Copy the halo on the device PSRC to Zxxxx_IN and put it in the PSRC copy on the host
#ifndef _PW_NOINTERM
IF (LX) THEN
   IF (.NOT. GWEST) THEN
   !$acc kernels async(IS_WEST)
   ZWEST_IN ( IIB:IIB+IHALO_1  ,    IJB:IJE  , : )  = PSRC( IIB:IIB+IHALO_1  ,  IJB:IJE  , : )
   !$acc end kernels
   !$acc update host(ZWEST_IN) async(IS_WEST)
   END IF
   IF (.NOT.GEAST) THEN
   !$acc kernels async(IS_EAST)
   ZEAST_IN ( IIE-IHALO_1:IIE  ,    IJB:IJE  , : )  = PSRC( IIE-IHALO_1:IIE  ,  IJB:IJE  , : )
   !$acc end kernels
   !$acc update host(ZEAST_IN) async(IS_EAST)
   ENDIF
END IF
IF (LY) THEN
   IF (.NOT.GSOUTH) THEN
   !$acc kernels async(IS_SOUTH)
   ZSOUTH_IN ( IIB:IIE  ,    IJB:IJB+IHALO_1  , : ) = PSRC( IIB:IIE  ,    IJB:IJB+IHALO_1  , : )
   !$acc end kernels
   !$acc update host(ZSOUTH_IN) async(IS_SOUTH)
   ENDIF
   IF (.NOT.GNORTH) THEN
   !$acc kernels async(IS_NORTH)
   ZNORTH_IN ( IIB:IIE  ,    IJE-IHALO_1:IJE  , : ) = PSRC( IIB:IIE  ,    IJE-IHALO_1:IJE  , : )
   !$acc end kernels
   !$acc update host(ZNORTH_IN) async(IS_NORTH)
   ENDIF
ENDIF
!$acc wait
IF (LX) THEN
   IF (.NOT. GWEST) THEN
   PSRC( IIB:IIB+IHALO_1  , IJB:IJE  , : )   = ZWEST_IN  ( IIB:IIB+IHALO_1   , IJB:IJE   , : )
   ENDIF
   IF (.NOT.GEAST) THEN 
   PSRC( IIE-IHALO_1:IIE  , IJB:IJE  , : )   = ZEAST_IN  ( IIE-IHALO_1:IIE   , IJB:IJE   , : )
   ENDIF
END IF
IF (LY) THEN
   IF (.NOT.GSOUTH) THEN
   PSRC( IIB:IIE  , IJB:IJB+IHALO_1  , : ) = ZSOUTH_IN ( IIB:IIE  ,    IJB:IJB+IHALO_1  , : )
   ENDIF
   IF (.NOT.GNORTH) THEN
   PSRC( IIB:IIE  , IJE-IHALO_1:IJE  , : ) = ZNORTH_IN ( IIB:IIE  ,    IJE-IHALO_1:IJE  , : )
   ENDIF
ENDIF
#else
IF (LX) THEN
   IF (.NOT. GWEST) THEN
   !$acc update host(PSRC( IIB:IIB+IHALO_1  , IJB:IJE  , : ))
   ENDIF
   IF (.NOT.GEAST) THEN
   !$acc update host(PSRC( IIE-IHALO_1:IIE  , IJB:IJE  , : ))
   ENDIF
END IF
IF (LY) THEN
   IF (.NOT.GSOUTH) THEN
   !$acc update host(PSRC( IIB:IIE  , IJB:IJB+IHALO_1  , : ))
   ENDIF
   IF (.NOT.GNORTH) THEN
   !$acc update host(PSRC( IIB:IIE  , IJE-IHALO_1:IJE  , : ))
   ENDIF
ENDIF
#endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

IF (LX .OR. LY) THEN
CALL ADD3DFIELD_ll( TZ_PSRC_ll, PSRC, 'GET_HALO::PSRC' )
CALL UPDATE_HALO_ll(TZ_PSRC_ll,IERROR, HDIR=HDIR )
CALL CLEANLIST_ll(TZ_PSRC_ll)
ELSE
!Necessary to allow comparisons/checks with standard GET_HALO
CALL MPPDB_CHECK3D(PSRC,"UPDATE_HALO_ll",PRECISION)
ENDIF

!Copy the halo on the host PSRC to Zxxxx_OUT and put it in the PSRC copy on the device
#ifndef _PW_NOINTERM
IF (LX) THEN
   IF (.NOT.GWEST) THEN
   ZWEST_OUT( 1:IIB-1  ,   IJB:IJE    , : )  = PSRC( 1:IIB-1  ,   IJB:IJE    , : )
   ENDIF
   IF (.NOT.GEAST) THEN
   ZEAST_OUT( IIE+1:IIU  ,   IJB:IJE    , : )  = PSRC( IIE+1:IIU  ,   IJB:IJE    , : )
   ENDIF
END IF
IF (LY) THEN
   IF (.NOT.GSOUTH) THEN
   ZSOUTH_OUT (  IIB:IIE     , 1:IJB-1  , : ) = PSRC(  IIB:IIE     , 1:IJB-1  , : )
   ENDIF
   IF (.NOT.GNORTH) THEN
   ZNORTH_OUT (  IIB:IIE     , IJE+1:IJU  , : ) = PSRC(  IIB:IIE     , IJE+1:IJU  , : )
   ENDIF
END IF
IF (LX) THEN
   IF (.NOT.GWEST) THEN
   !$acc update device(ZWEST_OUT) async(IS_WEST)
   !$acc kernels async(IS_WEST)
   PSRC( 1:IIB-1  ,      IJB:IJE      , : ) = ZWEST_OUT( 1:IIB-1  ,   IJB:IJE    , : )
   !$acc end kernels
   ENDIF
   IF (.NOT.GEAST) THEN
   !$acc update device(ZEAST_OUT) async(IS_EAST)
   !$acc kernels async(IS_EAST)
   PSRC( IIE+1:IIU  ,      IJB:IJE      , : ) = ZEAST_OUT( IIE+1:IIU  ,   IJB:IJE    , : )  
   !$acc end kernels
   ENDIF
END IF
IF (LY) THEN
   IF (.NOT.GSOUTH) THEN
   !$acc update device(ZSOUTH_OUT) async(IS_SOUTH)
   !$acc kernels async(IS_SOUTH)
   PSRC(      IIB:IIE       ,  1:IJB-1 , : ) = ZSOUTH_OUT(  IIB:IIE     , 1:IJB-1  , : )
   !$acc end kernels
   ENDIF
   IF (.NOT.GNORTH) THEN
   !$acc update device(ZNORTH_OUT) async(IS_NORTH)
   !$acc kernels async(IS_NORTH)
   PSRC(      IIB:IIE       , IJE+1:IJU , : ) = ZNORTH_OUT (  IIB:IIE     , IJE+1:IJU  , : )
   !$acc end kernels
   ENDIF
END IF
!$acc wait
#else
IF (LX) THEN
   IF (.NOT.GWEST) THEN
   !$acc update device(PSRC( 1:IIB-1  ,   IJB:IJE    , : ))
   ENDIF
   IF (.NOT.GEAST) THEN
   !$acc update device(PSRC( IIE+1:IIU  ,   IJB:IJE    , : ))
   ENDIF
END IF
IF (LY) THEN
   IF (.NOT.GSOUTH) THEN
   !$acc update device(PSRC(  IIB:IIE     , 1:IJB-1  , : ))
   ENDIF
   IF (.NOT.GNORTH) THEN
   !$acc update device(PSRC(  IIB:IIE     , IJE+1:IJU  , : ))
   ENDIF
END IF
#endif
!
END SUBROUTINE GET_HALO_D
#endif
!-----------------------------------------------------------------------
!
!
!     ####################################
      SUBROUTINE DEL_HALO2_ll(TPHALO2LIST)
!     ####################################
!
!!****  *DEL_HALO2_ll* delete the second layer of the halo
!!
!!
!!    Purpose
!!    -------
!       The purpose of this routine is to deallocate the 
!     TPHALO2LIST variable which contains the second layer of the
!     halo for each variable.
! 
!!    Implicit Arguments
!!    ------------------
!     Module MODD_ARGSLIST_ll
!       type HALO2LIST_ll
!!
!!    Reference
!!    ---------
! 
!!    Author
!!    ------
!     J. Escobar                 * LA - CNRS *
!
!     Modification :
!     -------------
!     Juan  11/03/2010 : Memory Leak add DEALLOCATE(TZHALO2LIST%HALO2)     
! 
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
  USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
!
  IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
  TYPE(HALO2LIST_ll), POINTER :: TPHALO2LIST ! list of HALO2_lls
!
!
!*       0.2   Declarations of local variables :
!
  TYPE(HALO2LIST_ll), POINTER :: TZHALO2LIST
!
!-------------------------------------------------------------------------------
!
!*       1.    Deallocate the list of HALO2_lls
!
  TZHALO2LIST => TPHALO2LIST
!
  DO WHILE(ASSOCIATED(TZHALO2LIST))
!
    TPHALO2LIST => TZHALO2LIST%NEXT
    DEALLOCATE(TZHALO2LIST%HALO2%WEST)
    DEALLOCATE(TZHALO2LIST%HALO2%EAST)
    DEALLOCATE(TZHALO2LIST%HALO2%SOUTH)
    DEALLOCATE(TZHALO2LIST%HALO2%NORTH)
    DEALLOCATE(TZHALO2LIST%HALO2)
    DEALLOCATE(TZHALO2LIST)
    TZHALO2LIST => TPHALO2LIST
!
  ENDDO
!
!-------------------------------------------------------------------------------
!
      END SUBROUTINE DEL_HALO2_ll
