!MNH_LIC Copyright 1994-2018 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
MODULE MODI_ICE4_COMPUTE_PDF
INTERFACE
SUBROUTINE ICE4_COMPUTE_PDF(KSIZE, HSUBG_AUCV, HSUBG_PR_PDF, &
                            PRHODREF, PRCT, PCF, PSIGMA_RC,&
                            PHLC_HCF, PHLC_LCF, PHLC_HRC, PHLC_LRC, PRF)
IMPLICIT NONE
INTEGER,                INTENT(IN)  :: KSIZE
CHARACTER(LEN=4),       INTENT(IN)  :: HSUBG_AUCV     ! Kind of Subgrid autoconversion method
CHARACTER(LEN=80),      INTENT(IN)  :: HSUBG_PR_PDF   ! pdf for subgrid precipitation
REAL, DIMENSION(KSIZE), INTENT(IN)  :: PRHODREF   ! Reference density
REAL, DIMENSION(KSIZE), INTENT(IN)  :: PRCT       ! Cloud water m.r. at t
REAL, DIMENSION(KSIZE), INTENT(IN)  :: PCF        ! Cloud fraction
REAL, DIMENSION(KSIZE), INTENT(IN)  :: PSIGMA_RC  ! Standard deviation of rc at time t
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PHLC_HCF   ! HLCLOUDS : fraction of High Cloud Fraction in grid
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PHLC_LCF   ! HLCLOUDS : fraction of Low  Cloud Fraction in grid
                                                  !    note that PCF = PHLC_HCF + PHLC_LCF
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PHLC_HRC   ! HLCLOUDS : LWC that is High LWC in grid
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PHLC_LRC   ! HLCLOUDS : LWC that is Low  LWC in grid
                                                  !    note that PRC = PHLC_HRC + PHLC_LRC
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PRF        ! Rain fraction
END SUBROUTINE ICE4_COMPUTE_PDF
END INTERFACE
END MODULE MODI_ICE4_COMPUTE_PDF
SUBROUTINE ICE4_COMPUTE_PDF(KSIZE, HSUBG_AUCV, HSUBG_PR_PDF, &
                            PRHODREF, PRCT, PCF, PSIGMA_RC,&
                            PHLC_HCF, PHLC_LCF, PHLC_HRC, PHLC_LRC, PRF)
!!
!!**  PURPOSE
!!    -------
!!      Computes the pdf used to split cloud into high and low content parts
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the plitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!
!
!*      0. DECLARATIONS
!          ------------
!
!
USE MODD_RAIN_ICE_DESCR, ONLY: XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: XCRIAUTC
!
USE MODE_MSG
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER,                INTENT(IN)  :: KSIZE
CHARACTER(LEN=4),       INTENT(IN)  :: HSUBG_AUCV     ! Kind of Subgrid autoconversion method
CHARACTER(LEN=80),      INTENT(IN)  :: HSUBG_PR_PDF   ! pdf for subgrid precipitation
REAL, DIMENSION(KSIZE), INTENT(IN)  :: PRHODREF   ! Reference density
REAL, DIMENSION(KSIZE), INTENT(IN)  :: PRCT       ! Cloud water m.r. at t
REAL, DIMENSION(KSIZE), INTENT(IN)  :: PCF        ! Cloud fraction
REAL, DIMENSION(KSIZE), INTENT(IN)  :: PSIGMA_RC  ! Standard deviation of rc at time t
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PHLC_HCF   ! HLCLOUDS : fraction of High Cloud Fraction in grid
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PHLC_LCF   ! HLCLOUDS : fraction of Low  Cloud Fraction in grid
                                                  !    note that PCF = PHLC_HCF + PHLC_LCF
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PHLC_HRC   ! HLCLOUDS : LWC that is High LWC in grid
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PHLC_LRC   ! HLCLOUDS : LWC that is Low  LWC in grid
                                                  !    note that PRC = PHLC_HRC + PHLC_LRC
REAL, DIMENSION(KSIZE), INTENT(OUT) :: PRF        ! Rain fraction
!
!$acc declare present(KSIZE,HSUBG_AUCV,HSUBG_PR_PDF,PRHODREF,PRCT,PCF,PSIGMA_RC, &
!$acc&                PHLC_HCF,PHLC_LCF,PHLC_HRC,PHLC_LRC,PRF)
!
!*       0.2  declaration of local variables
!
INTEGER                         :: JI
LOGICAL, DIMENSION(KSIZE)       :: GWORK, GWORK2, GWORK3
REAL, DIMENSION(SIZE(PRHODREF)) :: ZRCRAUTC,      & !RC value to begin rain formation =XCRIAUTC/RHODREF
                                   ZHLC_RCMAX,    & !HLCLOUDS : maximum value for RC in distribution
                                   ZHLC_LRCLOCAL, & !HLCLOUDS : LWC that is Low  LWC local in LCF
                                   ZHLC_HRCLOCAL    !HLCLOUDS : LWC that is High LWC local in HCF
                                                    !    note that ZRC/CF = ZHLC_HRCLOCAL+ ZHLC_LRCLOCAL
                                                    !                     = PHLC_HRC/HCF+ PHLC_LRC/LCF
REAL :: ZCOEFFRCM
!
!$acc declare create(GWORK,GWORK2,GWORK3,ZRCRAUTC,ZHLC_RCMAX,ZHLC_LRCLOCAL,ZHLC_HRCLOCAL,ZCOEFFRCM)
!
!$acc declare copyin(XCRIAUTC,XRTMIN)
!
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK1D(PRHODREF,"ICE4_COMPUTE_PDF beg:PRHODREF",PRECISION)
  CALL MPPDB_CHECK1D(PRCT,"ICE4_COMPUTE_PDF beg:PRCT",PRECISION)
  CALL MPPDB_CHECK1D(PCF,"ICE4_COMPUTE_PDF beg:PCF",PRECISION)
  CALL MPPDB_CHECK1D(PSIGMA_RC,"ICE4_COMPUTE_PDF beg:PSIGMA_RC",PRECISION)
END IF
!
!Cloud water split between high and low content part is done according to autoconversion option
!$acc kernels
ZRCRAUTC(:)=XCRIAUTC/PRHODREF(:) ! Autoconversion rc threshold
!$acc end kernels
IF(HSUBG_AUCV=='NONE') THEN
#if 0
!PW: bug: crash at execution with PGI 18.10
!$acc kernels
  !Cloud water is entirely in low or high part
  GWORK(:)  = PRCT(:)>ZRCRAUTC(:)
  GWORK2(:) = PRCT(:)>XRTMIN(2)
  WHERE(GWORK(:))
    PHLC_HCF(:)=1.
    PHLC_LCF(:)=0.
    PHLC_HRC(:)=PRCT(:)
    PHLC_LRC(:)=0.
    PRF(:)     =1.
  ELSEWHERE(GWORK2(:))
    PHLC_HCF(:)=0.
    PHLC_LCF(:)=1.
    PHLC_HRC(:)=0.
    PHLC_LRC(:)=PRCT(:)
    PRF(:)     =0.
  ELSEWHERE
    PHLC_HCF(:)=0.
    PHLC_LCF(:)=0.
    PHLC_HRC(:)=0.
    PHLC_LRC(:)=0.
    PRF(:)     =0.
  END WHERE
!$acc end kernels
#else
!$acc kernels
  DO JI = 1,KSIZE
    IF (PRCT(JI)>ZRCRAUTC(JI)) THEN
      PHLC_HCF(JI)=1.
      PHLC_LCF(JI)=0.
      PHLC_HRC(JI)=PRCT(JI)
      PHLC_LRC(JI)=0.
      PRF(JI)     =1.
    ELSE IF (PRCT(JI)>XRTMIN(2)) THEN
      PHLC_HCF(JI)=0.
      PHLC_LCF(JI)=1.
      PHLC_HRC(JI)=0.
      PHLC_LRC(JI)=PRCT(JI)
      PRF(JI)     =0.
    ELSE
      PHLC_HCF(JI)=0.
      PHLC_LCF(JI)=0.
      PHLC_HRC(JI)=0.
      PHLC_LRC(JI)=0.
      PRF(JI)     =0.
    END IF
  END DO
!$acc end kernels
#endif
ELSEIF(HSUBG_AUCV=='CLFR') THEN
!$acc kernels
  !Cloud water is only in the cloudy part and entirely in low or high part
  GWORK3(:) = PCF(:) > 0.
  GWORK(:)  = GWORK3(:) .AND. PRCT(:)>ZRCRAUTC(:)*PCF(:)
  GWORK2(:) = GWORK3(:) .AND. PRCT(:)>XRTMIN(2)
  WHERE(GWORK(:))
    PHLC_HCF(:)=PCF(:)
    PHLC_LCF(:)=0.
    PHLC_HRC(:)=PRCT(:)
    PHLC_LRC(:)=0.
    PRF(:)     =PCF(:)
  ELSEWHERE(GWORK2(:))
    PHLC_HCF(:)=0.
    PHLC_LCF(:)=PCF(:)
    PHLC_HRC(:)=0.0
    PHLC_LRC(:)=PRCT(:)
    PRF(:)     =0.
  ELSEWHERE (GWORK3(:))
    PHLC_HCF(:)=0.
    PHLC_LCF(:)=0.
    PHLC_HRC(:)=0.
    PHLC_LRC(:)=0.
    PRF(:)     =0.
  ELSEWHERE
    PHLC_HCF(:)=0.
    PHLC_LCF(:)=0.
    PHLC_HRC(:)=0.
    PHLC_LRC(:)=0.
    PRF(:)     =0.
  END WHERE
!$acc end kernels

ELSEIF(HSUBG_AUCV=='PDF ') THEN
  !Cloud water is split between high and low part according to a PDF
  !    'HLCRECTPDF'    : rectangular PDF form
  !    'HLCTRIANGPDF'  : triangular PDF form
  !    'HLCQUADRAPDF'  : second order quadratic PDF form
  !    'HLCISOTRIPDF'  : isocele triangular PDF
  !    'SIGM'          : Redelsperger and Sommeria (1986)
  IF(HSUBG_PR_PDF=='SIGM') THEN
!$acc kernels
    ! Redelsperger and Sommeria (1986) but organised according to Turner (2011, 2012)
    GWORK(:)  = PRCT(:)>  ZRCRAUTC(:)+PSIGMA_RC(:)
    GWORK2(:) = PRCT(:)> (ZRCRAUTC(:)-PSIGMA_RC(:)) .AND. &
                PRCT(:)<=(ZRCRAUTC(:)+PSIGMA_RC(:))
    GWORK3(:) = PRCT(:)>XRTMIN(2) .AND. PCF(:)>0.
    WHERE (GWORK(:))
      PHLC_HCF(:)=1.
      PHLC_LCF(:)=0.
      PHLC_HRC(:)=PRCT(:)
      PHLC_LRC(:)=0.
      PRF(:)     =1.
    ELSEWHERE(GWORK2(:))
      PHLC_HCF(:)=(PRCT(:)+PSIGMA_RC(:)-ZRCRAUTC(:))/ &
                  &(2.*PSIGMA_RC(:))
      PHLC_LCF(:)=MAX(0., PCF(:)-PHLC_HCF(:))
      PHLC_HRC(:)=(PRCT(:)+PSIGMA_RC(:)-ZRCRAUTC(:))* &
                  &(PRCT(:)+PSIGMA_RC(:)+ZRCRAUTC(:))/ &
                  &(4.*PSIGMA_RC(:))
      PHLC_LRC(:)=MAX(0., PRCT(:)-PHLC_HRC(:))
      PRF(:)     =PHLC_HCF(:)
    ELSEWHERE(GWORK3(:))
      PHLC_HCF(:)=0.
      PHLC_LCF(:)=PCF(:)
      PHLC_HRC(:)=0.
      PHLC_LRC(:)=PRCT(:)
      PRF(:)     =0.
    ELSEWHERE
      PHLC_HCF(:)=0.
      PHLC_LCF(:)=0.
      PHLC_HRC(:)=0.
      PHLC_LRC(:)=0.
      PRF(:)     =0.
    END WHERE
   ! Turner (2011, 2012)
!$acc end kernels
  ELSEIF(HSUBG_PR_PDF=='HLCRECTPDF' .OR. HSUBG_PR_PDF=='HLCISOTRIPDF' .OR. &
         &HSUBG_PR_PDF=='HLCTRIANGPDF' .OR. HSUBG_PR_PDF=='HLCQUADRAPDF') THEN
    ! Calculate maximum value r_cM from PDF forms
    IF(HSUBG_PR_PDF=='HLCRECTPDF' .OR. HSUBG_PR_PDF=='HLCISOTRIPDF') THEN
      ZCOEFFRCM=2.
    ELSE IF(HSUBG_PR_PDF=='HLCTRIANGPDF') THEN
      ZCOEFFRCM=3.
    ELSE IF(HSUBG_PR_PDF=='HLCQUADRAPDF') THEN
      ZCOEFFRCM=4.
    END IF
!$acc update device(ZCOEFFRCM)
!$acc kernels
    GWORK(:) = PRCT(:)>0. .AND. PCF(:)>0.
    WHERE(GWORK(:))
      ZHLC_RCMAX(:)=ZCOEFFRCM*PRCT(:)/PCF(:)
    END WHERE
    ! Split available water and cloud fraction in two parts
    ! Calculate local mean values int he low and high parts for the 3 PDF forms:
    GWORK(:) = PRCT(:)>0. .AND. PCF(:)>0. .AND. ZHLC_RCMAX(:)>ZRCRAUTC(:)
    IF(HSUBG_PR_PDF=='HLCRECTPDF') THEN
      WHERE(GWORK(:))
        ZHLC_LRCLOCAL(:)=0.5*ZRCRAUTC(:)
        ZHLC_HRCLOCAL(:)=( ZHLC_RCMAX(:) + ZRCRAUTC(:))/2.0
      END WHERE
    ELSE IF(HSUBG_PR_PDF=='HLCTRIANGPDF') THEN
      WHERE(GWORK(:))
        ZHLC_LRCLOCAL(:)=( ZRCRAUTC(:) *(3.0 * ZHLC_RCMAX(:) - 2.0 * ZRCRAUTC(:) ) ) &
                        / (3.0 * (2.0 * ZHLC_RCMAX(:) - ZRCRAUTC(:)  ) )
        ZHLC_HRCLOCAL(:)=(ZHLC_RCMAX(:) + 2.0*ZRCRAUTC(:)) / 3.0
      END WHERE
    ELSE IF(HSUBG_PR_PDF=='HLCQUADRAPDF') THEN
      WHERE(GWORK(:))
#ifndef MNH_BITREP
        ZHLC_LRCLOCAL(:)=(3.0 *ZRCRAUTC(:)**3 - 8.0 *ZRCRAUTC(:)**2 * ZHLC_RCMAX(:) &
                        + 6.0*ZRCRAUTC(:) *ZHLC_RCMAX(:)**2 ) &
                        / &
                        (4.0* ZRCRAUTC(:)**2 -12.0*ZRCRAUTC(:) *ZHLC_RCMAX(:) &
                        + 12.0 * ZHLC_RCMAX(:)**2 )
#else
        ZHLC_LRCLOCAL(:)=(3.0 *BR_P3(ZRCRAUTC(:)) - 8.0 *BR_P2(ZRCRAUTC(:)) * ZHLC_RCMAX(:) &
                        + 6.0*ZRCRAUTC(:) *BR_P2(ZHLC_RCMAX(:)) ) &
                        / &
                        (4.0* BR_P2(ZRCRAUTC(:)) -12.0*ZRCRAUTC(:) *ZHLC_RCMAX(:) &
                        + 12.0 * BR_P2(ZHLC_RCMAX(:)) )
#endif
        ZHLC_HRCLOCAL(:)=(ZHLC_RCMAX(:) + 3.0*ZRCRAUTC(:))/4.0
      END WHERE
    ELSE IF(HSUBG_PR_PDF=='HLCISOTRIPDF') THEN
      GWORK2(:) = PRCT(:) <= ZRCRAUTC(:)*PCF(:)
      WHERE(GWORK(:))
        WHERE(GWORK2(:))
#ifndef MNH_BITREP
          ZHLC_LRCLOCAL(:)=( (ZHLC_RCMAX(:))**3 &
                          -(12.0 * (ZHLC_RCMAX(:))*(ZRCRAUTC(:))**2) &
                          +(8.0 * ZRCRAUTC(:)**3) ) &
                          /( (6.0 * (ZHLC_RCMAX(:))**2) &
                          -(24.0 * (ZHLC_RCMAX(:)) * ZRCRAUTC(:)) &
                          +(12.0 * ZRCRAUTC(:)**2) )
#else
          ZHLC_LRCLOCAL(:)=( BR_P3(ZHLC_RCMAX(:)) &
                          -(12.0 * (ZHLC_RCMAX(:))*BR_P2(ZRCRAUTC(:))) &
                          +(8.0 * BR_P3(ZRCRAUTC(:))) ) &
                          /( (6.0 * BR_P2(ZHLC_RCMAX(:))) &
                          -(24.0 * (ZHLC_RCMAX(:)) * ZRCRAUTC(:)) &
                          +(12.0 * BR_P2(ZRCRAUTC(:))) )
#endif
          ZHLC_HRCLOCAL(:)=( ZHLC_RCMAX(:) + 2.0 * ZRCRAUTC(:) )/3.0
        ELSEWHERE
          ZHLC_LRCLOCAL(:)=(2.0/3.0) * ZRCRAUTC(:)
#ifndef MNH_BITREP
          ZHLC_HRCLOCAL(:)=(3.0*ZHLC_RCMAX(:)**3 - 8.0*ZRCRAUTC(:)**3) &
                          / (6.0 * ZHLC_RCMAX(:)**2 - 12.0*ZRCRAUTC(:)**2)
#else
          ZHLC_HRCLOCAL(:)=(3.0*BR_P3(ZHLC_RCMAX(:)) - 8.0*BR_P3(ZRCRAUTC(:))) &
                          / (6.0 * BR_P2(ZHLC_RCMAX(:)) - 12.0*BR_P2(ZRCRAUTC(:)))
#endif
        END WHERE
      END WHERE
    END IF
    ! Compare r_cM  to r_cR to know if cloud water content is high enough to split in two parts or not
    GWORK2(:) = PRCT(:)>0. .AND. PCF(:)>0. .AND. ZHLC_RCMAX(:)<=ZRCRAUTC(:)
    WHERE(GWORK(:))
      ! Calculate final values for LCF and HCF:
      PHLC_LCF(:)=PCF(:) &
                    *(ZHLC_HRCLOCAL- &
                    (PRCT(:) / PCF(:))) &
                    / (ZHLC_HRCLOCAL-ZHLC_LRCLOCAL)
      PHLC_HCF(:)=MAX(0., PCF(:)-PHLC_LCF(:))
      !
      ! Calculate final values for LRC and HRC:
      PHLC_LRC(:)=ZHLC_LRCLOCAL*PHLC_LCF(:)
      PHLC_HRC(:)=MAX(0., PRCT(:)-PHLC_LRC(:))
    ELSEWHERE (GWORK2(:))
      ! Put all available cloud water and his fraction in the low part
      PHLC_LCF(:)=PCF(:)
      PHLC_HCF(:)=0.
      PHLC_LRC(:)=PRCT(:)
      PHLC_HRC(:)=0.
    ELSEWHERE
      PHLC_LCF(:)=0.
      PHLC_HCF(:)=0.
      PHLC_LRC(:)=0.
      PHLC_HRC(:)=0.
    END WHERE
     PRF(:)=PHLC_HCF(:) !Precipitation fraction
!$acc end kernels
   ELSE
     CALL PRINT_MSG(NVERB_FATAL,'GEN','ICE4_COMPUTE_PDF','wrong HSUBG_PR_PDF case')
   ENDIF
ELSE
  CALL PRINT_MSG(NVERB_FATAL,'GEN','ICE4_COMPUTE_PDF','wrong HSUBG_AUCV case')
ENDIF
!
IF (MPPDB_INITIALIZED) THEN
  !Check all OUT arrays
  CALL MPPDB_CHECK1D(PHLC_HCF,"ICE4_COMPUTE_PDF end:PHLC_HCF",PRECISION)
  CALL MPPDB_CHECK1D(PHLC_LCF,"ICE4_COMPUTE_PDF end:PHLC_LCF",PRECISION)
  CALL MPPDB_CHECK1D(PHLC_HRC,"ICE4_COMPUTE_PDF end:PHLC_HRC",PRECISION)
  CALL MPPDB_CHECK1D(PHLC_LRC,"ICE4_COMPUTE_PDF end:PHLC_LRC",PRECISION)
  CALL MPPDB_CHECK1D(PRF,"ICE4_COMPUTE_PDF end:PRF",PRECISION)
END IF
!
END SUBROUTINE ICE4_COMPUTE_PDF
