!MNH_LIC Copyright 1994-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_FAST_RG
INTERFACE
SUBROUTINE ICE4_FAST_RG(KSIZE, LDSOFT, LDCOMPUTE, KRR, &
                       &PRHODREF, PLVFACT, PLSFACT, PPRES, &
                       &PDV, PKA, PCJ, PCIT, &
                       &PLBDAR, PLBDAS, PLBDAG, &
                       &PT,  PRVT, PRCT, PRRT, PRIT, PRST, PRGT, &
                       &PRGSI, PRGSI_MR, &
                       &LDWETG, &
                       &PRICFRRG, PRRCFRIG, PRICFRR, PRCWETG, PRIWETG, PRRWETG, PRSWETG, &
                       &PRCDRYG, PRIDRYG, PRRDRYG, PRSDRYG, PRWETGH, PRWETGH_MR, PRGMLTR, &
                       &PRG_TEND, &
                       &PA_TH, PA_RC, PA_RR, PA_RI, PA_RS, PA_RG, PA_RH, PB_RG, PB_RH)
IMPLICIT NONE
INTEGER,                      INTENT(IN)    :: KSIZE
LOGICAL,                      INTENT(IN)    :: LDSOFT
LOGICAL, DIMENSION(KSIZE),    INTENT(IN)    :: LDCOMPUTE
INTEGER,                      INTENT(IN)    :: KRR      ! Number of moist variable
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PCIT     ! Pristine ice conc. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLBDAG   ! Slope parameter of the graupel   distribution
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRGT     ! Graupel m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRGSI    ! Graupel tendency by other processes
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRGSI_MR ! Graupel mr change by other processes
LOGICAL, DIMENSION(KSIZE),    INTENT(OUT)   :: LDWETG   ! True where graupel grows in wet mode
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PRICFRRG ! Rain contact freezing
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PRRCFRIG ! Rain contact freezing
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PRICFRR  ! Rain contact freezing
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRCWETG  ! Graupel wet growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRIWETG  ! Graupel wet growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRRWETG  ! Graupel wet growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRSWETG  ! Graupel wet growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRCDRYG  ! Graupel dry growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRIDRYG  ! Graupel dry growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRRDRYG  ! Graupel dry growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRSDRYG  ! Graupel dry growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRWETGH  ! Conversion of graupel into hail
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRWETGH_MR ! Conversion of graupel into hail, mr change
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PRGMLTR  ! Melting of the graupel
REAL, DIMENSION(KSIZE, 6),    INTENT(INOUT) :: PRG_TEND ! Individual tendencies
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RR
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RI
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RG
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RH
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RG
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RH
END SUBROUTINE ICE4_FAST_RG
END INTERFACE
END MODULE MODI_ICE4_FAST_RG
SUBROUTINE ICE4_FAST_RG(KSIZE, LDSOFT, LDCOMPUTE, KRR, &
                       &PRHODREF, PLVFACT, PLSFACT, PPRES, &
                       &PDV, PKA, PCJ, PCIT, &
                       &PLBDAR, PLBDAS, PLBDAG, &
                       &PT,  PRVT, PRCT, PRRT, PRIT, PRST, PRGT, &
                       &PRGSI, PRGSI_MR, &
                       &LDWETG, &
                       &PRICFRRG, PRRCFRIG, PRICFRR, PRCWETG, PRIWETG, PRRWETG, PRSWETG, &
                       &PRCDRYG, PRIDRYG, PRRDRYG, PRSDRYG, PRWETGH, PRWETGH_MR, PRGMLTR, &
                       &PRG_TEND, &
                       &PA_TH, PA_RC, PA_RR, PA_RI, PA_RS, PA_RG, PA_RH, PB_RG, PB_RH)
!!
!!**  PURPOSE
!!    -------
!!      Computes the fast rg processes
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!  P. Wautelet 26/04/2019: replace non-standard FLOAT function by REAL function
!  P. Wautelet 29/05/2019: remove PACK/UNPACK intrinsics (to get more performance and better OpenACC support)
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XALPI,XALPW,XBETAI,XBETAW,XGAMW,XCI,XCL,XCPV,XESTT,XGAMI,XLMTT,XLVTT,XMD,XMV,XRV,XTT
USE MODD_PARAM_ICE,      ONLY: LCRFLIMIT,LEVLIMIT,LNULLWETG,LWETGPOST
USE MODD_RAIN_ICE_DESCR, ONLY: XBS,XCEXVT,XCXG,XCXS,XDG,XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: NDRYLBDAG,NDRYLBDAR,NDRYLBDAS,X0DEPG,X1DEPG,XCOLEXIG,XCOLEXSG,XCOLIG,XCOLSG,XDRYINTP1G, &
                               XDRYINTP1R,XDRYINTP1S,XDRYINTP2G,XDRYINTP2R,XDRYINTP2S,XEX0DEPG,XEX1DEPG,XEXICFRR,      &
                               XEXRCFRI,XFCDRYG,XFIDRYG,XFRDRYG,XFSDRYG,XICFRR,XKER_RDRYG,XKER_SDRYG,XLBRDRYG1,        &
                               XLBRDRYG2,XLBRDRYG3,XLBSDRYG1,XLBSDRYG2,XLBSDRYG3,XRCFRI
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER,                      INTENT(IN)    :: KSIZE
LOGICAL,                      INTENT(IN)    :: LDSOFT
LOGICAL, DIMENSION(KSIZE),    INTENT(IN)    :: LDCOMPUTE
INTEGER,                      INTENT(IN)    :: KRR      ! Number of moist variable
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PCIT     ! Pristine ice conc. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLBDAG   ! Slope parameter of the graupel   distribution
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRGT     ! Graupel m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRGSI    ! Graupel tendency by other processes
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRGSI_MR ! Graupel mr change by other processes
LOGICAL, DIMENSION(KSIZE),    INTENT(OUT)   :: LDWETG   ! True where graupel grows in wet mode
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PRICFRRG ! Rain contact freezing
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PRRCFRIG ! Rain contact freezing
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PRICFRR  ! Rain contact freezing
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRCWETG  ! Graupel wet growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRIWETG  ! Graupel wet growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRRWETG  ! Graupel wet growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRSWETG  ! Graupel wet growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRCDRYG  ! Graupel dry growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRIDRYG  ! Graupel dry growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRRDRYG  ! Graupel dry growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRSDRYG  ! Graupel dry growth
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRWETGH  ! Conversion of graupel into hail
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRWETGH_MR ! Conversion of graupel into hail, mr change
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PRGMLTR  ! Melting of the graupel
REAL, DIMENSION(KSIZE, 6),    INTENT(INOUT) :: PRG_TEND ! Individual tendencies
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RR
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RI
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RG
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PA_RH
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RG
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RH
!
!$acc declare present(KSIZE,LDSOFT,LDCOMPUTE,KRR,PRHODREF,PLVFACT,PLSFACT,PPRES,PDV,PKA,PCJ,PCIT,PLBDAR,PLBDAS,PLBDAG,  &
!$acc&                PT,PRVT,PRCT,PRRT,PRIT,PRST,PRGT,PRGSI,PRGSI_MR,LDWETG,PRICFRRG,PRRCFRIG,PRICFRR,PRCWETG,PRIWETG, &
!$acc&                PRRWETG,PRSWETG,PRCDRYG,PRIDRYG,PRRDRYG,PRSDRYG,PRWETGH,PRWETGH_MR,PRGMLTR,PRG_TEND,              &
!$acc&                PA_TH,PA_RC,PA_RR,PA_RI,PA_RS,PA_RG,PA_RH,PB_RG,PB_RH)
!
!*       0.2  declaration of local variables
!
INTEGER, PARAMETER :: IRCDRYG=1, IRIDRYG=2, IRIWETG=3, IRSDRYG=4, IRSWETG=5, IRRDRYG=6
!
INTEGER                            :: IGDRY
INTEGER                            :: IDX, JJ
INTEGER                            :: IRR !Workaround of PGI bug with OpenACC (at least up to 18.10 version)
INTEGER, DIMENSION(SIZE(PRHODREF)) :: I1
INTEGER, DIMENSION(SIZE(PRHODREF)) :: IVEC1, IVEC2
LOGICAL                            :: GCRFLIMIT,GDSOFT,GEVLIMIT,GNULLWETG,GWETGPOST !Workaround of PGI bug with OpenACC (at least up to 18.10 version)
LOGICAL, DIMENSION(SIZE(PRHODREF)) :: GDRY, GLDRYG, GMASK
REAL, DIMENSION(SIZE(PRHODREF))    :: ZVEC1, ZVEC2, ZVEC3
REAL, DIMENSION(SIZE(PRHODREF))    :: ZZW,         &
                                      ZRDRYG_INIT, & !Initial dry growth rate of the graupeln
                                      ZRWETG_INIT    !Initial wet growth rate of the graupeln
!
!$acc declare create(IGDRY,I1,IVEC1,IVEC2,GDRY,GLDRYG,GMASK,ZVEC1,ZVEC2,ZVEC3,ZZW,ZRDRYG_INIT,ZRWETG_INIT)
!
!$acc declare copyin(XALPI,XALPW,XBETAI,XBETAW,XGAMW,XCI,XCL,XCPV,XESTT,XGAMI,XLMTT,XLVTT,XMD,XMV,XRV,XTT,   &
!$acc&               LCRFLIMIT,LEVLIMIT,LNULLWETG,LWETGPOST,                                                 &
!$acc&               XBS,XCEXVT,XCXG,XCXS,XDG,XRTMIN,                                                        &
!$acc&               NDRYLBDAG,NDRYLBDAR,NDRYLBDAS,X0DEPG,X1DEPG,XCOLEXIG,XCOLEXSG,XCOLIG,XCOLSG,XDRYINTP1G, &
!$acc&               XDRYINTP1R,XDRYINTP1S,XDRYINTP2G,XDRYINTP2R,XDRYINTP2S,XEX0DEPG,XEX1DEPG,XEXICFRR,      &
!$acc&               XEXRCFRI,XFCDRYG,XFIDRYG,XFRDRYG,XFSDRYG,XICFRR,XKER_RDRYG,XKER_SDRYG,XLBRDRYG1,        &
!$acc&               XLBRDRYG2,XLBRDRYG3,XLBSDRYG1,XLBSDRYG2,XLBSDRYG3,XRCFRI)
!
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK1D_LOG(LDCOMPUTE,"ICE4_FAST_RG beg:LDCOMPUTE")
  CALL MPPDB_CHECK1D(PRHODREF,"ICE4_FAST_RG beg:PRHODREF",PRECISION)
  CALL MPPDB_CHECK1D(PLVFACT,"ICE4_FAST_RG beg:PLVFACT",PRECISION)
  CALL MPPDB_CHECK1D(PLSFACT,"ICE4_FAST_RG beg:PLSFACT",PRECISION)
  CALL MPPDB_CHECK1D(PPRES,"ICE4_FAST_RG beg:PPRES",PRECISION)
  CALL MPPDB_CHECK1D(PDV,"ICE4_FAST_RG beg:PDV",PRECISION)
  CALL MPPDB_CHECK1D(PKA,"ICE4_FAST_RG beg:PKA",PRECISION)
  CALL MPPDB_CHECK1D(PCJ,"ICE4_FAST_RG beg:PCJ",PRECISION)
  CALL MPPDB_CHECK1D(PCIT,"ICE4_FAST_RG beg:PCIT",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAR,"ICE4_FAST_RG beg:PLBDAR",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAS,"ICE4_FAST_RG beg:PLBDAS",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAG,"ICE4_FAST_RG beg:PLBDAG",PRECISION)
  CALL MPPDB_CHECK1D(PT,"ICE4_FAST_RG beg:PT",PRECISION)
  CALL MPPDB_CHECK1D(PRVT,"ICE4_FAST_RG beg:PRVT",PRECISION)
  CALL MPPDB_CHECK1D(PRCT,"ICE4_FAST_RG beg:PRCT",PRECISION)
  CALL MPPDB_CHECK1D(PRRT,"ICE4_FAST_RG beg:PRRT",PRECISION)
  CALL MPPDB_CHECK1D(PRIT,"ICE4_FAST_RG beg:PRIT",PRECISION)
  CALL MPPDB_CHECK1D(PRST,"ICE4_FAST_RG beg:PRST",PRECISION)
  CALL MPPDB_CHECK1D(PRGT,"ICE4_FAST_RG beg:PRGT",PRECISION)
  CALL MPPDB_CHECK1D(PRGSI,"ICE4_FAST_RG beg:PRGSI",PRECISION)
  CALL MPPDB_CHECK1D(PRGSI_MR,"ICE4_FAST_RG beg:PRGSI_MR",PRECISION)
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRICFRRG,"ICE4_FAST_RG beg:PRICFRRG",PRECISION)
  CALL MPPDB_CHECK1D(PRRCFRIG,"ICE4_FAST_RG beg:PRRCFRIG",PRECISION)
  CALL MPPDB_CHECK1D(PRICFRR,"ICE4_FAST_RG beg:PRICFRR",PRECISION)
  CALL MPPDB_CHECK1D(PRGMLTR,"ICE4_FAST_RG beg:PRGMLTR",PRECISION)
  CALL MPPDB_CHECK2D(PRG_TEND,"ICE4_FAST_RG beg:PRG_TEND",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RG beg:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RG beg:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_FAST_RG beg:PA_RR",PRECISION)
  CALL MPPDB_CHECK1D(PA_RI,"ICE4_FAST_RG beg:PA_RI",PRECISION)
  CALL MPPDB_CHECK1D(PA_RS,"ICE4_FAST_RG beg:PA_RS",PRECISION)
  CALL MPPDB_CHECK1D(PA_RG,"ICE4_FAST_RG beg:PA_RG",PRECISION)
  CALL MPPDB_CHECK1D(PA_RH,"ICE4_FAST_RG beg:PA_RH",PRECISION)
  CALL MPPDB_CHECK1D(PB_RG,"ICE4_FAST_RG beg:PB_RG",PRECISION)
  CALL MPPDB_CHECK1D(PB_RH,"ICE4_FAST_RG beg:PB_RH",PRECISION)
END IF
!
!Workaround of PGI bug with OpenACC (at least up to 18.10 version)
GCRFLIMIT = LCRFLIMIT
GEVLIMIT  = LEVLIMIT
GNULLWETG = LNULLWETG
GWETGPOST = LWETGPOST
! !$acc kernels
!PW: bug?: for the moment I take the value from the CPU as OK. GPU value seems to be lost when this subroutine is called
GDSOFT    = LDSOFT
! !$acc end kernels
!$acc kernels
IRR       = KRR
!$acc end kernels
!
!*       6.1    rain contact freezing
!
!$acc kernels
GMASK(:)=PRIT(:)>XRTMIN(4) .AND. PRRT(:)>XRTMIN(3) .AND. LDCOMPUTE(:)
IF(GDSOFT) THEN
  WHERE(.NOT. GMASK(:))
    PRICFRRG(:)=0.
    PRRCFRIG(:)=0.
    PRICFRR(:)=0.
  ENDWHERE
ELSE
  PRICFRRG(:)=0.
  PRRCFRIG(:)=0.
  PRICFRR(:)=0.
  WHERE(GMASK(:))
#ifndef MNH_BITREP
    PRICFRRG(:) = XICFRR*PRIT(:)                & ! RICFRRG
                                 *PLBDAR(:)**XEXICFRR    &
                                 *PRHODREF(:)**(-XCEXVT)
    PRRCFRIG(:) = XRCFRI*PCIT(:)                & ! RRCFRIG
                                 * PLBDAR(:)**XEXRCFRI    &
                                 * PRHODREF(:)**(-XCEXVT-1.)
#else
    PRICFRRG(:) = XICFRR*PRIT(:)                & ! RICFRRG
                                 *BR_POW(PLBDAR(:),XEXICFRR)    &
                                 *BR_POW(PRHODREF(:),-XCEXVT)
    PRRCFRIG(:) = XRCFRI*PCIT(:)                & ! RRCFRIG
                                 * BR_POW(PLBDAR(:),XEXRCFRI)    &
                                 * BR_POW(PRHODREF(:),-XCEXVT-1.)
#endif
  END WHERE
  ZZW(:)=1.
  IF(GCRFLIMIT) THEN
    WHERE(GMASK(:))
      !Comparison between heat to be released (to freeze rain) and heat sink (rain and ice temperature change)
      !ZZW is the proportion of process that can take place
      ZZW(:) = MAX(0., MIN(1., (PRICFRRG(:)*XCI+PRRCFRIG(:)*XCL)*(XTT-PT(:)) / &
                               MAX(1.E-20, XLVTT*PRRCFRIG(:))))
    ENDWHERE
  ENDIF
  PRRCFRIG(:) = ZZW(:) * PRRCFRIG(:) !Part of rain that can be freezed
  PRICFRR(:) = (1-ZZW(:)) * PRICFRRG(:) !Part of collected pristine ice converted to rain
  PRICFRRG(:) = ZZW(:) * PRICFRRG(:) !Part of collected pristine ice that lead to graupel
ENDIF
PA_RI(:) = PA_RI(:) - PRICFRRG(:) - PRICFRR(:)
PA_RR(:) = PA_RR(:) - PRRCFRIG(:) + PRICFRR(:)
PA_RG(:) = PA_RG(:) + PRICFRRG(:) + PRRCFRIG(:)
PA_TH(:) = PA_TH(:) + (PRRCFRIG(:) - PRICFRR(:))*(PLSFACT(:)-PLVFACT(:))
!
!
!*       6.3    compute the graupel growth
!
! Wet and dry collection of rc and ri on graupel
GMASK(:)=PRGT(:)>XRTMIN(6) .AND. PRCT(:)>XRTMIN(2) .AND. LDCOMPUTE(:)
IF(GDSOFT) THEN
  WHERE(.NOT. GMASK(:))
    PRG_TEND(:, IRCDRYG)=0.
  END WHERE
ELSE
  PRG_TEND(:, IRCDRYG)=0.
  WHERE(GMASK(:))
#ifndef MNH_BITREP
    ZZW(:)=PLBDAG(:)**(XCXG-XDG-2.) * PRHODREF(:)**(-XCEXVT)
#else
    ZZW(:)=BR_POW(PLBDAG(:),XCXG-XDG-2.) * BR_POW(PRHODREF(:),-XCEXVT)
#endif
    PRG_TEND(:, IRCDRYG)=XFCDRYG * PRCT(:) * ZZW(:)
  END WHERE
ENDIF
GMASK(:)=PRGT(:)>XRTMIN(6) .AND. PRIT(:)>XRTMIN(4) .AND. LDCOMPUTE(:)
IF(GDSOFT) THEN
  WHERE(.NOT. GMASK(:))
    PRG_TEND(:, IRIDRYG)=0.
    PRG_TEND(:, IRIWETG)=0.
  END WHERE
ELSE
  PRG_TEND(:, IRIDRYG)=0.
  PRG_TEND(:, IRIWETG)=0.
  WHERE(GMASK(:))
#ifndef MNH_BITREP
    ZZW(:)=PLBDAG(:)**(XCXG-XDG-2.) * PRHODREF(:)**(-XCEXVT)
    PRG_TEND(:, IRIDRYG)=XFIDRYG*EXP(XCOLEXIG*(PT(:)-XTT))*PRIT(:)*ZZW(:)
    PRG_TEND(:, IRIWETG)=PRG_TEND(:, IRIDRYG) / (XCOLIG*EXP(XCOLEXIG*(PT(:)-XTT)))
#else
    ZZW(:)=BR_POW(PLBDAG(:),XCXG-XDG-2.) * BR_POW(PRHODREF(:),-XCEXVT)
    PRG_TEND(:, IRIDRYG)=XFIDRYG*BR_EXP(XCOLEXIG*(PT(:)-XTT))*PRIT(:)*ZZW(:)
    PRG_TEND(:, IRIWETG)=PRG_TEND(:, IRIDRYG) / (XCOLIG*BR_EXP(XCOLEXIG*(PT(:)-XTT)))
#endif
  END WHERE
ENDIF

! Wet and dry collection of rs on graupel (6.2.1)
IGDRY = 0
!$acc loop private(IDX) independent
DO JJ = 1, SIZE(GDRY)
  IF (PRST(JJ)>XRTMIN(5) .AND. PRGT(JJ)>XRTMIN(6) .AND. LDCOMPUTE(JJ)) THEN
!$acc atomic capture
    IGDRY = IGDRY + 1
    IDX = IGDRY
!$acc end atomic
    I1(IDX) = JJ
    GDRY(JJ) = .TRUE.
  ELSE
    GDRY(JJ) = .FALSE.
  END IF
END DO
!$acc end kernels
!PW: BUG: this is necessary to get correct results (PGI 18.10)
! !$acc update self(GDRY,IGDRY)
!$acc update self(IGDRY)
IF(JJ==-999) print *,'PW: IGDRY=',IGDRY,COUNT(GDRY)
! !$acc kernels
IF(GDSOFT) THEN
!$acc kernels
  WHERE(.NOT. GDRY(:))
    PRG_TEND(:, IRSDRYG)=0.
    PRG_TEND(:, IRSWETG)=0.
  END WHERE
!$acc end kernels
ELSE
!$acc kernels
  PRG_TEND(:, IRSDRYG)=0.
  PRG_TEND(:, IRSWETG)=0.
!$acc end kernels
  IF(IGDRY>0)THEN
!$acc kernels
    !
    !*       6.2.3  select the (PLBDAG,PLBDAS) couplet
    !
    DO JJ = 1, IGDRY
      ZVEC1(JJ) = PLBDAG(I1(JJ))
      ZVEC2(JJ) = PLBDAS(I1(JJ))
    END DO
    !
    !*       6.2.4  find the next lower indice for the PLBDAG and for the PLBDAS
    !               in the geometrical set of (Lbda_g,Lbda_s) couplet use to
    !               tabulate the SDRYG-kernel
    !
    ZVEC1(1:IGDRY)=MAX(1.00001, MIN(REAL(NDRYLBDAG)-0.00001,           &
#ifndef MNH_BITREP
                          XDRYINTP1G*LOG(ZVEC1(1:IGDRY))+XDRYINTP2G))
#else
                          XDRYINTP1G*BR_LOG(ZVEC1(1:IGDRY))+XDRYINTP2G))
#endif
    IVEC1(1:IGDRY)=INT(ZVEC1(1:IGDRY) )
    ZVEC1(1:IGDRY)=ZVEC1(1:IGDRY)-REAL(IVEC1(1:IGDRY))
    !
    ZVEC2(1:IGDRY)=MAX(1.00001, MIN( REAL(NDRYLBDAS)-0.00001,           &
#ifndef MNH_BITREP
                          XDRYINTP1S*LOG(ZVEC2(1:IGDRY))+XDRYINTP2S))
#else
                          XDRYINTP1S*BR_LOG(ZVEC2(1:IGDRY))+XDRYINTP2S))
#endif
    IVEC2(1:IGDRY)=INT(ZVEC2(1:IGDRY))
    ZVEC2(1:IGDRY)=ZVEC2(1:IGDRY)-REAL(IVEC2(1:IGDRY))
    !
    !*       6.2.5  perform the bilinear interpolation of the normalized
    !               SDRYG-kernel
    !
    DO JJ=1, IGDRY
      ZVEC3(JJ) =  (  XKER_SDRYG(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_SDRYG(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         * ZVEC1(JJ) &
                 - (  XKER_SDRYG(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_SDRYG(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         *(ZVEC1(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
    DO JJ = 1, IGDRY
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    !
    WHERE(GDRY(:))
#ifndef MNH_BITREP
      PRG_TEND(:, IRSWETG)=XFSDRYG*ZZW(:)                         & ! RSDRYG
                                    / XCOLSG &
                  *(PLBDAS(:)**(XCXS-XBS))*( PLBDAG(:)**XCXG )    &
                  *(PRHODREF(:)**(-XCEXVT-1.))                    &
                       *( XLBSDRYG1/( PLBDAG(:)**2              ) + &
                          XLBSDRYG2/( PLBDAG(:)   * PLBDAS(:)   ) + &
                          XLBSDRYG3/(               PLBDAS(:)**2))
      PRG_TEND(:, IRSDRYG)=PRG_TEND(:, IRSWETG)*XCOLSG*EXP(XCOLEXSG*(PT(:)-XTT))
#else
      PRG_TEND(:, IRSWETG)=XFSDRYG*ZZW(:)                            & ! RSDRYG
                                    / XCOLSG                         &
                  *BR_POW(PLBDAS(:),XCXS-XBS)*BR_POW(PLBDAG(:),XCXG) &
                  *BR_POW(PRHODREF(:),-XCEXVT-1.)                    &
                       *( XLBSDRYG1/  BR_P2(PLBDAG(:))             + &
                          XLBSDRYG2/( PLBDAG(:)   * PLBDAS(:)   )  + &
                          XLBSDRYG3/              BR_P2(PLBDAS(:)) )
      PRG_TEND(:, IRSDRYG)=PRG_TEND(:, IRSWETG)*XCOLSG*BR_EXP(XCOLEXSG*(PT(:)-XTT))
#endif
    END WHERE
!$acc end kernels
  ENDIF
ENDIF
!$acc kernels
!
!*       6.2.6  accretion of raindrops on the graupeln
!
IGDRY = 0
!$acc loop private(IDX) independent
DO JJ = 1, SIZE(GDRY)
  IF (PRRT(JJ)>XRTMIN(3) .AND. PRGT(JJ)>XRTMIN(6) .AND. LDCOMPUTE(JJ)) THEN
!$acc atomic capture
    IGDRY = IGDRY + 1
    IDX = IGDRY
!$acc end atomic
    I1(IDX) = JJ
    GDRY(JJ) = .TRUE.
  ELSE
    GDRY(JJ) = .FALSE.
  END IF
END DO
!$acc end kernels
!PW: BUG: this is necessary to get correct results (PGI 18.10)
! !$acc update self(GDRY,IGDRY)
!$acc update self(IGDRY)
IF(JJ==-999) print *,'PW: IGDRY=',IGDRY,COUNT(GDRY)
! !$acc kernels
IF(GDSOFT) THEN
!$acc kernels
  WHERE(.NOT. GDRY(:))
    PRG_TEND(:, IRRDRYG)=0.
  END WHERE
!$acc end kernels
ELSE
!$acc kernels
  PRG_TEND(:, IRRDRYG)=0.
!$acc end kernels
  !
  IF(IGDRY>0) THEN
!$acc kernels
    !
    !*       6.2.8  select the (PLBDAG,PLBDAR) couplet
    !
    DO JJ = 1, IGDRY
      ZVEC1(JJ) = PLBDAG(I1(JJ))
      ZVEC2(JJ) = PLBDAR(I1(JJ))
    END DO
    !
    !*       6.2.9  find the next lower indice for the PLBDAG and for the PLBDAR
    !               in the geometrical set of (Lbda_g,Lbda_r) couplet use to
    !               tabulate the RDRYG-kernel
    !
    ZVEC1(1:IGDRY)=MAX(1.00001, MIN( REAL(NDRYLBDAG)-0.00001,           &
#ifndef MNH_BITREP
                          XDRYINTP1G*LOG(ZVEC1(1:IGDRY))+XDRYINTP2G))
#else
                          XDRYINTP1G*BR_LOG(ZVEC1(1:IGDRY))+XDRYINTP2G))
#endif
    IVEC1(1:IGDRY)=INT(ZVEC1(1:IGDRY))
    ZVEC1(1:IGDRY)=ZVEC1(1:IGDRY)-REAL(IVEC1(1:IGDRY))
    !
    ZVEC2(1:IGDRY)=MAX(1.00001, MIN( REAL(NDRYLBDAR)-0.00001,           &
#ifndef MNH_BITREP
                          XDRYINTP1R*LOG(ZVEC2(1:IGDRY))+XDRYINTP2R))
#else
                          XDRYINTP1R*BR_LOG(ZVEC2(1:IGDRY))+XDRYINTP2R))
#endif
    IVEC2(1:IGDRY)=INT(ZVEC2(1:IGDRY))
    ZVEC2(1:IGDRY)=ZVEC2(1:IGDRY)-REAL(IVEC2(1:IGDRY))
    !
    !*       6.2.10 perform the bilinear interpolation of the normalized
    !               RDRYG-kernel
    !
    DO JJ=1, IGDRY
      ZVEC3(JJ)= (  XKER_RDRYG(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RDRYG(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                                  * ZVEC1(JJ) &
                 - (  XKER_RDRYG(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RDRYG(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         *(ZVEC1(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
    DO JJ = 1, IGDRY
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    !
    WHERE(GDRY(:))
#ifndef MNH_BITREP
      PRG_TEND(:, IRRDRYG) = XFRDRYG*ZZW(:)                    & ! RRDRYG
                        *( PLBDAR(:)**(-4) )*( PLBDAG(:)**XCXG ) &
                               *( PRHODREF(:)**(-XCEXVT-1.) )   &
                    *( XLBRDRYG1/( PLBDAG(:)**2              ) + &
                       XLBRDRYG2/( PLBDAG(:)   * PLBDAR(:)   ) + &
                       XLBRDRYG3/(               PLBDAR(:)**2) )
#else
      PRG_TEND(:, IRRDRYG) = XFRDRYG*ZZW(:)                         & ! RRDRYG
                        / BR_P4(PLBDAR(:)) * BR_POW(PLBDAG(:),XCXG) &
                               * BR_POW(PRHODREF(:),-XCEXVT-1.)     &
                    *( XLBRDRYG1/ BR_P2(PLBDAG(:))                + &
                       XLBRDRYG2/( PLBDAG(:)   * PLBDAR(:)   )    + &
                       XLBRDRYG3/               BR_P2(PLBDAR(:))  )
#endif
    END WHERE
!$acc end kernels
  ENDIF
ENDIF
!$acc kernels

ZRDRYG_INIT(:)=PRG_TEND(:, IRCDRYG)+PRG_TEND(:, IRIDRYG)+PRG_TEND(:, IRSDRYG)+PRG_TEND(:, IRRDRYG)

!Freezing rate
ZRWETG_INIT(:)=0.
GMASK(:)=PRGT(:)>XRTMIN(6) .AND. LDCOMPUTE(:)
WHERE(GMASK(:))
  ZRWETG_INIT(:)=PRVT(:)*PPRES(:)/((XMV/XMD)+PRVT(:)) ! Vapor pressure
END WHERE
IF(GEVLIMIT) THEN
  WHERE(GMASK(:))
#ifndef MNH_BITREP
    ZRWETG_INIT(:)=MIN(ZRWETG_INIT(:), EXP(XALPI-XBETAI/PT(:)-XGAMI*ALOG(PT(:)))) ! min(ev, es_i(T))
#else
    ZRWETG_INIT(:)=MIN(ZRWETG_INIT(:), BR_EXP(XALPI-XBETAI/PT(:)-XGAMI*BR_LOG(PT(:)))) ! min(ev, es_i(T))
#endif
  END WHERE
ENDIF
WHERE(GMASK(:))
  ZRWETG_INIT(:)=PKA(:)*(XTT-PT(:)) +                              &
           (PDV(:)*(XLVTT+(XCPV-XCL)*(PT(:)-XTT)) &
                         *(XESTT-ZRWETG_INIT(:))/(XRV*PT(:))           )
  ZRWETG_INIT(:)=MAX(0.,                                               &
#ifndef MNH_BITREP
             (ZRWETG_INIT(:) * ( X0DEPG*       PLBDAG(:)**XEX0DEPG +     &
                         X1DEPG*PCJ(:)*PLBDAG(:)**XEX1DEPG ) +   &
#else
             (ZRWETG_INIT(:) * ( X0DEPG*       BR_POW(PLBDAG(:),XEX0DEPG) +     &
                         X1DEPG*PCJ(:)*BR_POW(PLBDAG(:),XEX1DEPG) ) +   &
#endif
             (PRG_TEND(:, IRIWETG)+PRG_TEND(:, IRSWETG) ) *                            &
             (PRHODREF(:)*(XLMTT+(XCI-XCL)*(XTT-PT(:)))   ) ) / &
                        ( PRHODREF(:)*(XLMTT-XCL*(XTT-PT(:))) )   )
  !We must agregate, at least, the cold species
  ZRWETG_INIT(:)=MAX(ZRWETG_INIT(:), PRG_TEND(:, IRIWETG)+PRG_TEND(:, IRSWETG))
END WHERE

!Growth mode
LDWETG(:)=GMASK(:) .AND. &
         &MAX(0., ZRDRYG_INIT(:)-PRG_TEND(:, IRIDRYG)-PRG_TEND(:, IRSDRYG))>= &
         &MAX(0., ZRWETG_INIT(:)-PRG_TEND(:, IRIWETG)-PRG_TEND(:, IRSWETG))
IF(GNULLWETG) THEN
  LDWETG(:)=LDWETG(:) .AND. ZRDRYG_INIT(:)>0.
ELSE
  LDWETG(:)=LDWETG(:) .AND. ZRWETG_INIT(:)>0.
ENDIF
IF(.NOT. GWETGPOST) LDWETG(:)=LDWETG(:) .AND. PT(:)<XTT

GLDRYG(:)=GMASK(:) .AND. PT(:)<XTT .AND. ZRDRYG_INIT(:)>0. .AND. &
    &MAX(0., ZRDRYG_INIT(:)-PRG_TEND(:, IRIDRYG)-PRG_TEND(:, IRSDRYG))<&
    &MAX(0., ZRWETG_INIT(:)-PRG_TEND(:, IRIWETG)-PRG_TEND(:, IRSWETG))

! Part of ZRWETG to be converted into hail
! Graupel can be produced by other processes instantaneously (inducing a mixing ratio change, PRGSI_MR) or
! as a tendency (PRWETGH)
PRWETGH(:)=0.
PRWETGH_MR(:)=0.
IF(IRR==7) THEN
  WHERE(LDWETG(:))
    !assume a linear percent of conversion of produced graupel into hail
    PRWETGH(:)=(MAX(0., PRGSI(:)+PRICFRRG(:)+PRRCFRIG(:))+ZRWETG_INIT(:))*ZRDRYG_INIT(:)/(ZRWETG_INIT(:)+ZRDRYG_INIT(:))
    PRWETGH_MR(:)=MAX(0., PRGSI_MR(:))*ZRDRYG_INIT(:)/(ZRWETG_INIT(:)+ZRDRYG_INIT(:))
  END WHERE
ENDIF

PRCWETG(:)=0.
PRIWETG(:)=0.
PRSWETG(:)=0.
PRRWETG(:)=0.
WHERE(LDWETG(:))
  !Aggregated minus collected
  PRRWETG(:)=-(PRG_TEND(:, IRIWETG)+PRG_TEND(:, IRSWETG)+PRG_TEND(:, IRCDRYG)-ZRWETG_INIT(:))
  PRCWETG(:)=PRG_TEND(:, IRCDRYG)
  PRIWETG(:)=PRG_TEND(:, IRIWETG)
  PRSWETG(:)=PRG_TEND(:, IRSWETG)
END WHERE
PRCDRYG(:)=0.
PRIDRYG(:)=0.
PRRDRYG(:)=0.
PRSDRYG(:)=0.
WHERE(GLDRYG(:))
  PRCDRYG(:)=PRG_TEND(:, IRCDRYG)
  PRRDRYG(:)=PRG_TEND(:, IRRDRYG)
  PRIDRYG(:)=PRG_TEND(:, IRIDRYG)
  PRSDRYG(:)=PRG_TEND(:, IRSDRYG)
END WHERE
PA_RC(:) = PA_RC(:) - PRCWETG(:)
PA_RI(:) = PA_RI(:) - PRIWETG(:)
PA_RS(:) = PA_RS(:) - PRSWETG(:)
PA_RG(:) = PA_RG(:) + PRCWETG(:) + PRIWETG(:) + PRSWETG(:) + PRRWETG(:)
PA_RR(:) = PA_RR(:) - PRRWETG(:)
PA_TH(:) = PA_TH(:) + (PRCWETG(:) + PRRWETG(:))*(PLSFACT(:)-PLVFACT(:))
PA_RG(:) = PA_RG(:) - PRWETGH(:)
PA_RH(:) = PA_RH(:) + PRWETGH(:)
PB_RG(:) = PB_RG(:) - PRWETGH_MR(:)
PB_RH(:) = PB_RH(:) + PRWETGH_MR(:)
PA_RC(:) = PA_RC(:) - PRCDRYG(:)
PA_RI(:) = PA_RI(:) - PRIDRYG(:)
PA_RS(:) = PA_RS(:) - PRSDRYG(:)
PA_RR(:) = PA_RR(:) - PRRDRYG(:)
PA_RG(:) = PA_RG(:) + PRCDRYG(:) + PRIDRYG(:) + PRSDRYG(:) + PRRDRYG(:)
PA_TH(:) = PA_TH(:) + (PRCDRYG(:)+PRRDRYG(:))*(PLSFACT(:)-PLVFACT(:))

!
!*       6.5    Melting of the graupeln
!
GMASK(:)=PRGT(:)>XRTMIN(6) .AND. PT(:)>XTT .AND. LDCOMPUTE(:)
IF(GDSOFT) THEN
  WHERE(.NOT. GMASK(:))
    PRGMLTR(:) = 0.
  END WHERE
ELSE
  PRGMLTR(:) = 0.
  WHERE(GMASK(:))
    PRGMLTR(:) = PRVT(:)*PPRES(:)/((XMV/XMD)+PRVT(:)) ! Vapor pressure
  END WHERE
  IF(GEVLIMIT) THEN
    WHERE(GMASK(:))
#ifndef MNH_BITREP
      PRGMLTR(:)=MIN(PRGMLTR(:), EXP(XALPW-XBETAW/PT(:)-XGAMW*ALOG(PT(:)))) ! min(ev, es_w(T))
#else
      PRGMLTR(:)=MIN(PRGMLTR(:), BR_EXP(XALPW-XBETAW/PT(:)-XGAMW*BR_LOG(PT(:)))) ! min(ev, es_w(T))
#endif
    END WHERE
  ENDIF
  WHERE(GMASK(:))
    PRGMLTR(:) =  PKA(:)*(XTT-PT(:)) +                                 &
               ( PDV(:)*(XLVTT + ( XCPV - XCL ) * ( PT(:) - XTT )) &
                           *(XESTT-PRGMLTR(:))/(XRV*PT(:))             )
  END WHERE
  WHERE(GMASK(:))
    !
    ! compute RGMLTR
    !
    PRGMLTR(:)  = MAX( 0.0,( -PRGMLTR(:) *                     &
#ifndef MNH_BITREP
                           ( X0DEPG*       PLBDAG(:)**XEX0DEPG +     &
                             X1DEPG*PCJ(:)*PLBDAG(:)**XEX1DEPG ) -   &
#else
                           ( X0DEPG*       BR_POW(PLBDAG(:),XEX0DEPG) +     &
                             X1DEPG*PCJ(:)*BR_POW(PLBDAG(:),XEX1DEPG) ) -   &
#endif
                         ( PRG_TEND(:, IRCDRYG)+PRG_TEND(:, IRRDRYG) ) *       &
                               ( PRHODREF(:)*XCL*(XTT-PT(:))) ) /    &
                                             ( PRHODREF(:)*XLMTT ) )
  END WHERE
ENDIF
PA_RR(:) = PA_RR(:) + PRGMLTR(:)
PA_RG(:) = PA_RG(:) - PRGMLTR(:)
PA_TH(:) = PA_TH(:) - PRGMLTR(:)*(PLSFACT(:)-PLVFACT(:))
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRICFRRG,"ICE4_FAST_RG end:PRICFRRG",PRECISION)
  CALL MPPDB_CHECK1D(PRRCFRIG,"ICE4_FAST_RG end:PRRCFRIG",PRECISION)
  CALL MPPDB_CHECK1D(PRICFRR,"ICE4_FAST_RG end:PRICFRR",PRECISION)
  CALL MPPDB_CHECK1D(PRGMLTR,"ICE4_FAST_RG end:PRGMLTR",PRECISION)
  CALL MPPDB_CHECK2D(PRG_TEND,"ICE4_FAST_RG end:PRG_TEND",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RG end:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RG end:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_FAST_RG end:PA_RR",PRECISION)
  CALL MPPDB_CHECK1D(PA_RI,"ICE4_FAST_RG end:PA_RI",PRECISION)
  CALL MPPDB_CHECK1D(PA_RS,"ICE4_FAST_RG end:PA_RS",PRECISION)
  CALL MPPDB_CHECK1D(PA_RG,"ICE4_FAST_RG end:PA_RG",PRECISION)
  CALL MPPDB_CHECK1D(PA_RH,"ICE4_FAST_RG end:PA_RH",PRECISION)
  CALL MPPDB_CHECK1D(PB_RG,"ICE4_FAST_RG end:PB_RG",PRECISION)
  CALL MPPDB_CHECK1D(PB_RH,"ICE4_FAST_RG end:PB_RH",PRECISION)
  !Check all OUT arrays
  CALL MPPDB_CHECK1D_LOG(LDWETG,"ICE4_FAST_RG end:LDWETG")
  CALL MPPDB_CHECK1D(PRCWETG,"ICE4_FAST_RG end:PRCWETG",PRECISION)
  CALL MPPDB_CHECK1D(PRIWETG,"ICE4_FAST_RG end:PRIWETG",PRECISION)
  CALL MPPDB_CHECK1D(PRRWETG,"ICE4_FAST_RG end:PRRWETG",PRECISION)
  CALL MPPDB_CHECK1D(PRSWETG,"ICE4_FAST_RG end:PRSWETG",PRECISION)
  CALL MPPDB_CHECK1D(PRCDRYG,"ICE4_FAST_RG end:PRCDRYG",PRECISION)
  CALL MPPDB_CHECK1D(PRIDRYG,"ICE4_FAST_RG end:PRIDRYG",PRECISION)
  CALL MPPDB_CHECK1D(PRRDRYG,"ICE4_FAST_RG end:PRRDRYG",PRECISION)
  CALL MPPDB_CHECK1D(PRSDRYG,"ICE4_FAST_RG end:PRSDRYG",PRECISION)
  CALL MPPDB_CHECK1D(PRWETGH,"ICE4_FAST_RG end:PRWETGH",PRECISION)
  CALL MPPDB_CHECK1D(PRWETGH_MR,"ICE4_FAST_RG end:PRWETGH_MR",PRECISION)
END IF
!
END SUBROUTINE ICE4_FAST_RG
