!MNH_LIC Copyright 1994-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
MODULE MODI_ICE4_RIMLTC
INTERFACE
SUBROUTINE ICE4_RIMLTC(KSIZE, LDSOFT, LDCOMPUTE, &
                       &PEXN, PLVFACT, PLSFACT, &
                       &PT, &
                       &PTHT, PRIT, &
                       &PRIMLTC_MR, PB_TH, PB_RC, PB_RI)
IMPLICIT NONE
INTEGER, INTENT(IN) :: KSIZE
LOGICAL,                  INTENT(IN)    :: LDSOFT
LOGICAL, DIMENSION(KSIZE),INTENT(IN)    :: LDCOMPUTE
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PEXN     ! Exner function
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PTHT     ! Theta at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRIT     ! Cloud ice at t
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRIMLTC_MR ! Mixing ratio change due to cloud ice melting
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_TH
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RC
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RI
END SUBROUTINE ICE4_RIMLTC
END INTERFACE
END MODULE MODI_ICE4_RIMLTC
SUBROUTINE ICE4_RIMLTC(KSIZE, LDSOFT, LDCOMPUTE, &
                       &PEXN, PLVFACT, PLSFACT, &
                       &PT, &
                       &PTHT, PRIT, &
                       &PRIMLTC_MR, PB_TH, PB_RC, PB_RI)
!!
!!**  PURPOSE
!!    -------
!!      Computes the RIMLTC process
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,       ONLY: XTT
USE MODD_PARAM_ICE, ONLY: LFEEDBACKT
!
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER,                      INTENT(IN)    :: KSIZE
LOGICAL,                      INTENT(IN)    :: LDSOFT
LOGICAL, DIMENSION(KSIZE),    INTENT(IN)    :: LDCOMPUTE
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PEXN     ! Exner function
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PTHT     ! Theta at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRIT     ! Cloud ice at t
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRIMLTC_MR ! Mixing ratio change due to cloud ice melting
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_TH
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RC
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RI
!
!$acc declare present(KSIZE,LDSOFT,LDCOMPUTE,PEXN,PLVFACT,PLSFACT,PT, &
!$acc&                PTHT,PRIT,PRIMLTC_MR,PB_TH,PB_RC,PB_RI)
!
!*       0.2  declaration of local variables
!
LOGICAL :: GDSOFT, GFEEDBACKT !Workaround of PGI bug with OpenACC (at least up to 18.10 version)
LOGICAL, DIMENSION(KSIZE) :: GMASK
!$acc declare create(GMASK)
!
!$acc declare copyin(XTT,LFEEDBACKT)
!-------------------------------------------------------------------------------
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PEXN,"ICE4_RIMLTC beg:PEXN")
  CALL MPPDB_CHECK(PLVFACT,"ICE4_RIMLTC beg:PLVFACT")
  CALL MPPDB_CHECK(PLSFACT,"ICE4_RIMLTC beg:PLSFACT")
  CALL MPPDB_CHECK(PT,"ICE4_RIMLTC beg:PT")
  CALL MPPDB_CHECK(PTHT,"ICE4_RIMLTC beg:PTHT")
  CALL MPPDB_CHECK(PRIT,"ICE4_RIMLTC beg:PRIT")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PB_TH,"ICE4_RIMLTC beg:PB_TH")
  CALL MPPDB_CHECK(PB_RC,"ICE4_RIMLTC beg:PB_RC")
  CALL MPPDB_CHECK(PB_RI,"ICE4_RIMLTC beg:PB_RI")
END IF
!
!*       7.1    cloud ice melting
!
! !$acc kernels
!PW: bug?: for the moment I take the value from the CPU as OK. GPU value seems to be lost when this subroutine is called
GDSOFT=LDSOFT
GFEEDBACKT=LFEEDBACKT !Workaround of PGI bug with OpenACC (at least up to 18.10 version)
! !$acc end kernels
!$acc kernels
PRIMLTC_MR(:)=0.
IF(.NOT. GDSOFT) THEN
  GMASK(:)=PRIT(:)>0. .AND. PT(:)>XTT .AND. LDCOMPUTE(:)
  WHERE(GMASK(:))
    PRIMLTC_MR(:)=PRIT(:)
  END WHERE

  IF(GFEEDBACKT) THEN
    !Limitation due to 0 crossing of temperature
    WHERE(GMASK(:))
      PRIMLTC_MR(:)=MIN(PRIMLTC_MR(:), MAX(0., (PTHT(:)-XTT/PEXN(:)) / (PLSFACT(:)-PLVFACT(:))))
    END WHERE
  ENDIF
ENDIF
PB_RC(:) = PB_RC(:) + PRIMLTC_MR(:)
PB_RI(:) = PB_RI(:) - PRIMLTC_MR(:)
PB_TH(:) = PB_TH(:) - PRIMLTC_MR(:)*(PLSFACT(:)-PLVFACT(:))
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PB_TH,"ICE4_RIMLTC end:PB_TH")
  CALL MPPDB_CHECK(PB_RC,"ICE4_RIMLTC end:PB_RC")
  CALL MPPDB_CHECK(PB_RI,"ICE4_RIMLTC end:PB_RI")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRIMLTC_MR,"ICE4_RIMLTC end:PRIMLTC_MR")
END IF
!
END SUBROUTINE ICE4_RIMLTC
