!MNH_LIC Copyright 1994-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_RSRIMCG_OLD
INTERFACE
SUBROUTINE ICE4_RSRIMCG_OLD(KSIZE, ODSOFT, ODCOMPUTE, &
                           &PRHODREF, &
                           &PLBDAS, &
                           &PT, PRCT, PRST, &
                           &PRSRIMCG_MR, PB_RS, PB_RG)
IMPLICIT NONE
INTEGER, INTENT(IN) :: KSIZE
LOGICAL,                      INTENT(IN)    :: ODSOFT
LOGICAL, DIMENSION(KSIZE),    INTENT(IN)    :: ODCOMPUTE
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRSRIMCG_MR ! Mr change due to cloud droplet riming of the aggregates
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RS
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RG
END SUBROUTINE ICE4_RSRIMCG_OLD
END INTERFACE
END MODULE MODI_ICE4_RSRIMCG_OLD
SUBROUTINE ICE4_RSRIMCG_OLD(KSIZE, ODSOFT, ODCOMPUTE, &
                           &PRHODREF, &
                           &PLBDAS, &
                           &PT, PRCT, PRST, &
                           &PRSRIMCG_MR, PB_RS, PB_RG)
!!
!!**  PURPOSE
!!    -------
!!      Computes the riming-conversion of the large sized aggregates into graupel
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!  P. Wautelet 26/04/2019: replace non-standard FLOAT function by REAL function
!  P. Wautelet 29/05/2019: remove PACK/UNPACK intrinsics (to get more performance and better OpenACC support)
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XTT
USE MODD_PARAM_ICE,      ONLY: CSNOWRIMING
USE MODD_RAIN_ICE_DESCR, ONLY: XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: NGAMINC,XEXSRIMCG,XGAMINC_RIM2,XRIMINTP1,XRIMINTP2,XSRIMCG
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER, INTENT(IN) :: KSIZE
LOGICAL,                      INTENT(IN)    :: ODSOFT
LOGICAL, DIMENSION(KSIZE),    INTENT(IN)    :: ODCOMPUTE
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(KSIZE),       INTENT(OUT)   :: PRSRIMCG_MR ! Mr change due to cloud droplet riming of the aggregates
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RS
REAL, DIMENSION(KSIZE),       INTENT(INOUT) :: PB_RG
!
!$acc declare present(KSIZE,ODSOFT,ODCOMPUTE,PRHODREF,PLBDAS, &
!$acc&                PT,PRCT,PRST,PRSRIMCG_MR,PB_RS,PB_RG)
!
!*       0.2  declaration of local variables
!
INTEGER                            :: IDX, JL
INTEGER                            :: IGRIM
INTEGER, DIMENSION(SIZE(PRHODREF)) :: IVEC1, IVEC2
LOGICAL                            :: GDSOFT !Workaround of PGI bug with OpenACC (at least up to 18.10 version)
LOGICAL, DIMENSION(SIZE(PRHODREF)) :: GRIM
REAL, DIMENSION(SIZE(PRHODREF))    :: ZVEC1, ZVEC2
REAL, DIMENSION(SIZE(PRHODREF))    :: ZZW
!
!$acc declare create(IGRIM, IVEC1, IVEC2, GRIM, ZVEC1, ZVEC2, ZZW)
!
!$acc declare copyin(XTT,CSNOWRIMING,XRTMIN,NGAMINC,XEXSRIMCG,XGAMINC_RIM2,XRIMINTP1,XRIMINTP2,XSRIMCG)
!
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(ODCOMPUTE,"ICE4_RSRIMCG_OLD beg:ODCOMPUTE")
  CALL MPPDB_CHECK(PRHODREF,"ICE4_RSRIMCG_OLD beg:PRHODREF")
  CALL MPPDB_CHECK(PLBDAS,"ICE4_RSRIMCG_OLD beg:PLBDAS")
  CALL MPPDB_CHECK(PT,"ICE4_RSRIMCG_OLD beg:PT")
  CALL MPPDB_CHECK(PRCT,"ICE4_RSRIMCG_OLD beg:PRCT")
  CALL MPPDB_CHECK(PRST,"ICE4_RSRIMCG_OLD beg:PRST")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PB_RS,"ICE4_RSRIMCG_OLD beg:PB_RS")
  CALL MPPDB_CHECK(PB_RG,"ICE4_RSRIMCG_OLD beg:PB_RG")
END IF
!
!-------------------------------------------------------------------------------
!
!*       5.1    cloud droplet riming of the aggregates
!
!$acc kernels
!PW: bug?: for the moment I take the value from the CPU as OK. GPU value seems to be lost when this subroutine is called
GDSOFT=ODSOFT !Workaround of PGI bug with OpenACC (at least up to 18.10 version)
!$acc end kernels
!$acc kernels
PRSRIMCG_MR(:)=0.
!
IF(.NOT. GDSOFT) THEN
  IGRIM = 0
  GRIM(:) = .FALSE.
!$acc loop private(IDX) independent
  DO JL = 1, SIZE(GRIM)
    IF ( PRCT(JL)>XRTMIN(2) .AND. PRST(JL)>XRTMIN(5) .AND. ODCOMPUTE(JL) .AND. PT(JL)<XTT ) THEN
!$acc atomic capture
      IGRIM = IGRIM + 1
      IDX = IGRIM
!$acc end atomic
      IVEC1(IDX) = Jl
      GRIM(JL) = .TRUE.
    END IF
  END DO
  !
  IF(IGRIM>0 .AND. CSNOWRIMING=='OLD ') THEN
    !
    !        5.1.1  select the PLBDAS
    !
    DO JL = 1, IGRIM
      ZVEC1(JL) = PLBDAS(IVEC1(JL))
    END DO
    !
    !        5.1.2  find the next lower indice for the PLBDAS in the geometrical
    !               set of Lbda_s used to tabulate some moments of the incomplete
    !               gamma function
    !
    ZVEC2(1:IGRIM) = MAX( 1.00001, MIN( REAL(NGAMINC)-0.00001,           &
#ifndef MNH_BITREP
                          XRIMINTP1 * LOG( ZVEC1(1:IGRIM) ) + XRIMINTP2 ) )
#else
                          XRIMINTP1 * BR_LOG( ZVEC1(1:IGRIM) ) + XRIMINTP2 ) )
#endif
    IVEC2(1:IGRIM) = INT( ZVEC2(1:IGRIM) )
    ZVEC2(1:IGRIM) = ZVEC2(1:IGRIM) - REAL( IVEC2(1:IGRIM) )

    !
    !        5.1.5  perform the linear interpolation of the normalized
    !               "XBS"-moment of the incomplete gamma function (XGAMINC_RIM2)
    !
    ZVEC1(1:IGRIM) =  XGAMINC_RIM2( IVEC2(1:IGRIM)+1 )* ZVEC2(1:IGRIM)      &
                    - XGAMINC_RIM2( IVEC2(1:IGRIM)   )*(ZVEC2(1:IGRIM) - 1.0)
    ZZW(:) = 0.
    DO JL = 1, IGRIM
      ZZW(IVEC1(JL)) = ZVEC1(JL)
    END DO
    !
    !        5.1.6  riming-conversion of the large sized aggregates into graupeln
    !
    !
    WHERE(GRIM(:))
#ifndef MNH_BITREP
      PRSRIMCG_MR(:) = XSRIMCG * PLBDAS(:)**XEXSRIMCG   & ! RSRIMCG
#else
      PRSRIMCG_MR(:) = XSRIMCG * BR_POW(PLBDAS(:),XEXSRIMCG)   & ! RSRIMCG
#endif
                               * (1.0 - ZZW(:) )/PRHODREF(:)
    END WHERE
    PRSRIMCG_MR(:)=MIN(PRST(:), PRSRIMCG_MR(:))
  END IF
ENDIF
PB_RS(:) = PB_RS(:) - PRSRIMCG_MR(:)
PB_RG(:) = PB_RG(:) + PRSRIMCG_MR(:)
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PB_RS,"ICE4_RSRIMCG_OLD end:PB_RS")
  CALL MPPDB_CHECK(PB_RG,"ICE4_RSRIMCG_OLD end:PB_RG")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRSRIMCG_MR,"ICE4_RSRIMCG_OLD end:PRSRIMCG_MR")
END IF
!
END SUBROUTINE ICE4_RSRIMCG_OLD
