!MNH_LIC Copyright 1995-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
! Modifications:
!  P. Wautelet 25/02/2019: split rain_ice (cleaner and easier to maintain/debug)
!  P. Wautelet 26/04/2019: replace non-standard FLOAT function by REAL function
!  P. Wautelet 03/06/2019: remove PACK/UNPACK intrinsics (to get more performance and better OpenACC support)
!  P. Wautelet 05/06/2019: optimisations
!-----------------------------------------------------------------
MODULE MODE_RAIN_ICE_FAST_RG

  IMPLICIT NONE

  PRIVATE

  PUBLIC :: RAIN_ICE_FAST_RG

CONTAINS

SUBROUTINE RAIN_ICE_FAST_RG(KRR, OMICRO, PRHODREF, PRVT, PRCT, PRRT, PRIT, PRST, PRGT, PCIT, &
                            PRHODJ, PPRES, PZT, PLBDAR, PLBDAS, PLBDAG, PLSFACT, PLVFACT, &
                            PCJ, PKA, PDV, PRHODJ3D, PTHS3D, PRCS, PRRS, PRIS, PRSS, PRGS, PRHS, PTHS, &
                            PUSW, PRDRYG, PRWETG)

!
!*      0. DECLARATIONS
!          ------------
!
use MODD_BUDGET,         only: LBUDGET_RC, LBUDGET_RG, LBUDGET_RH, LBUDGET_RI, LBUDGET_RR, LBUDGET_RS, LBUDGET_TH
use MODD_CST,            only: XCI, XCL, XCPV, XESTT, XLMTT, XLVTT, XMD, XMV, XRV, XTT
use MODD_RAIN_ICE_DESCR, only: XBS, XCEXVT, XCXG, XCXS, XDG, XRTMIN
use MODD_RAIN_ICE_PARAM, only: NDRYLBDAG, NDRYLBDAR, NDRYLBDAS, X0DEPG, X1DEPG, XCOLEXIG, XCOLEXSG, XCOLIG, XCOLSG, XDRYINTP1G, &
                               XDRYINTP1R, XDRYINTP1S, XDRYINTP2G, XDRYINTP2R, XDRYINTP2S, XEX0DEPG, XEX1DEPG, XEXICFRR,        &
                               XEXRCFRI, XFCDRYG, XFIDRYG, XFRDRYG, XFSDRYG, XICFRR, XKER_RDRYG, XKER_SDRYG, XLBRDRYG1,         &
                               XLBRDRYG2, XLBRDRYG3, XLBSDRYG1, XLBSDRYG2, XLBSDRYG3, XRCFRI
!
use mode_mppdb
#ifndef _OPENACC
use mode_tools,                        only: Countjv
#else
use mode_tools,                        only: Countjv_device
#endif
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
use MODI_BUDGET
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER,                    INTENT(IN)    :: KRR      ! Number of moist variables
LOGICAL,  DIMENSION(:,:,:), intent(in)    :: OMICRO   ! Test where to compute all processes
REAL,     DIMENSION(:),     intent(in)    :: PRHODREF ! RHO Dry REFerence
REAL,     DIMENSION(:),     intent(in)    :: PRVT     ! Water vapor m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRCT     ! Cloud water m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRRT     ! Rain water m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRIT     ! Pristine ice m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRST     ! Snow/aggregate m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRGT     ! Graupel m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PCIT     ! Pristine ice conc. at t
REAL,     DIMENSION(:),     intent(in)    :: PRHODJ   ! RHO times Jacobian
REAL,     DIMENSION(:),     intent(in)    :: PPRES    ! Pressure
REAL,     DIMENSION(:),     intent(in)    :: PZT      ! Temperature
REAL,     DIMENSION(:),     intent(in)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL,     DIMENSION(:),     intent(in)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL,     DIMENSION(:),     intent(in)    :: PLBDAG   ! Slope parameter of the graupel   distribution
REAL,     DIMENSION(:),     intent(in)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL,     DIMENSION(:),     intent(in)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL,     DIMENSION(:),     intent(in)    :: PCJ      ! Function to compute the ventilation coefficient
REAL,     DIMENSION(:),     intent(in)    :: PKA      ! Thermal conductivity of the air
REAL,     DIMENSION(:),     intent(in)    :: PDV      ! Diffusivity of water vapor in the air
REAL,     DIMENSION(:,:,:), INTENT(IN)    :: PRHODJ3D ! Dry density * Jacobian
REAL,     DIMENSION(:,:,:), INTENT(IN)    :: PTHS3D   ! Theta source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRCS     ! Cloud water m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRRS     ! Rain water m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRIS     ! Pristine ice m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRSS     ! Snow/aggregate m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRGS     ! Graupel m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRHS     ! Hail m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PTHS     ! Theta source
REAL,     DIMENSION(:),     intent(inout) :: PUSW     ! Undersaturation over water
REAL,     DIMENSION(:),     intent(out)   :: PRDRYG   ! Dry growth rate of the graupeln
REAL,     DIMENSION(:),     intent(out)   :: PRWETG   ! Wet growth rate of the graupeln
!
! IN variables
!
!$acc declare present(OMICRO, PRHODREF, PRVT, PRCT, PRRT, PRIT, PRST, PRGT, PCIT,   &
!$acc &               PRHODJ, PPRES, PZT, PLBDAR, PLBDAS, PLBDAG, PLSFACT, PLVFACT, &
!$acc &               PCJ, PKA, PDV, PRHODJ3D, PTHS3D)                              &
!
! INOUT variables
!
!$acc &       present(PRCS, PRRS, PRIS, PRSS, PRGS, PRHS, PTHS, PUSW)               &
!
! OUT variables
!
!$acc &       present(PRDRYG, PRWETG)
!
!*       0.2  declaration of local variables
!
INTEGER                              :: IGDRY
INTEGER                              :: JJ, JL
INTEGER, DIMENSION(:),   ALLOCATABLE :: I1
INTEGER, DIMENSION(:),   ALLOCATABLE :: IVEC1, IVEC2      ! Vectors of indices for interpolations
LOGICAL, DIMENSION(:),   ALLOCATABLE :: GWORK
REAL,    DIMENSION(:),   ALLOCATABLE :: ZZW               ! Work array
REAL,    DIMENSION(:),   ALLOCATABLE :: ZVEC1,ZVEC2,ZVEC3 ! Work vectors for interpolations
REAL,    DIMENSION(:),   ALLOCATABLE :: ZVECLBDAG, ZVECLBDAR, ZVECLBDAS
REAL,    DIMENSION(:,:), ALLOCATABLE :: ZZW1              ! Work arrays
!
!$acc declare device_resident(I1, IVEC1, IVEC2, GWORK, ZZW, ZVEC1, ZVEC2, ZVEC3, &
!$acc &                       ZVECLBDAG, ZVECLBDAR, ZVECLBDAS, ZZW1)
!
!$acc declare copyin(XKER_RDRYG, XKER_SDRYG)
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(OMICRO,"RAIN_ICE_FAST_RG beg:OMICRO")
  CALL MPPDB_CHECK(PRHODREF,"RAIN_ICE_FAST_RG beg:PRHODREF")
  CALL MPPDB_CHECK(PRVT,"RAIN_ICE_FAST_RG beg:PRVT")
  CALL MPPDB_CHECK(PRCT,"RAIN_ICE_FAST_RG beg:PRCT")
  CALL MPPDB_CHECK(PRRT,"RAIN_ICE_FAST_RG beg:PRRT")
  CALL MPPDB_CHECK(PRIT,"RAIN_ICE_FAST_RG beg:PRIT")
  CALL MPPDB_CHECK(PRST,"RAIN_ICE_FAST_RG beg:PRST")
  CALL MPPDB_CHECK(PRGT,"RAIN_ICE_FAST_RG beg:PRGT")
  CALL MPPDB_CHECK(PCIT,"RAIN_ICE_FAST_RG beg:PCIT")
  CALL MPPDB_CHECK(PRHODJ,"RAIN_ICE_FAST_RG beg:PRHODJ")
  CALL MPPDB_CHECK(PPRES,"RAIN_ICE_FAST_RG beg:PPRES")
  CALL MPPDB_CHECK(PZT,"RAIN_ICE_FAST_RG beg:PZT")
  CALL MPPDB_CHECK(PLBDAR,"RAIN_ICE_FAST_RG beg:PLBDAR")
  CALL MPPDB_CHECK(PLBDAS,"RAIN_ICE_FAST_RG beg:PLBDAS")
  CALL MPPDB_CHECK(PLBDAG,"RAIN_ICE_FAST_RG beg:PLBDAG")
  CALL MPPDB_CHECK(PLSFACT,"RAIN_ICE_FAST_RG beg:PLSFACT")
  CALL MPPDB_CHECK(PLVFACT,"RAIN_ICE_FAST_RG beg:PLVFACT")
  CALL MPPDB_CHECK(PCJ,"RAIN_ICE_FAST_RG beg:PCJ")
  CALL MPPDB_CHECK(PKA,"RAIN_ICE_FAST_RG beg:PKA")
  CALL MPPDB_CHECK(PDV,"RAIN_ICE_FAST_RG beg:PDV")
  CALL MPPDB_CHECK(PRHODJ3D,"RAIN_ICE_FAST_RG beg:PRHODJ3D")
  CALL MPPDB_CHECK(PTHS3D,"RAIN_ICE_FAST_RG beg:PTHS3D")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRCS,"RAIN_ICE_FAST_RG beg:PRCS")
  CALL MPPDB_CHECK(PRRS,"RAIN_ICE_FAST_RG beg:PRRS")
  CALL MPPDB_CHECK(PRIS,"RAIN_ICE_FAST_RG beg:PRIS")
  CALL MPPDB_CHECK(PRSS,"RAIN_ICE_FAST_RG beg:PRSS")
  CALL MPPDB_CHECK(PRGS,"RAIN_ICE_FAST_RG beg:PRGS")
  CALL MPPDB_CHECK(PRHS,"RAIN_ICE_FAST_RG beg:PRHS")
  CALL MPPDB_CHECK(PTHS,"RAIN_ICE_FAST_RG beg:PTHS")
  CALL MPPDB_CHECK(PUSW,"RAIN_ICE_FAST_RG beg:PUSW")
END IF
!
ALLOCATE( I1   (size(PRHODREF)) )
ALLOCATE( GWORK(size(PRHODREF)) )
ALLOCATE( ZZW  (size(PRHODREF)) )
ALLOCATE( ZZW1 (size(PRHODREF),7) )
!
!*       6.1    rain contact freezing
!
!$acc kernels
  GWORK(:) = PRIT(:)>XRTMIN(4) .AND. PRRT(:)>XRTMIN(3) .AND. PRIS(:)>0.0 .AND. PRRS(:)>0.0
  WHERE( GWORK(:) )
#ifndef MNH_BITREP
    ZZW1(:,3) = MIN( PRIS(:),XICFRR * PRIT(:)                & ! RICFRRG
                                    * PLBDAR(:)**XEXICFRR    &
                                    * PRHODREF(:)**(-XCEXVT) )
    ZZW1(:,4) = MIN( PRRS(:),XRCFRI * PCIT(:)                & ! RRCFRIG
                                    * PLBDAR(:)**XEXRCFRI    &
                                    * PRHODREF(:)**(-XCEXVT-1.) )
#else
    ZZW1(:,3) = MIN( PRIS(:),XICFRR * PRIT(:)                     & ! RICFRRG
                                    * BR_POW(PLBDAR(:),XEXICFRR)  &
                                    * BR_POW(PRHODREF(:),-XCEXVT) )
    ZZW1(:,4) = MIN( PRRS(:),XRCFRI * PCIT(:)                     & ! RRCFRIG
                                    * BR_POW(PLBDAR(:),XEXRCFRI)  &
                                    * BR_POW(PRHODREF(:),-XCEXVT-1.) )
#endif
    PRIS(:) = PRIS(:) - ZZW1(:,3)
    PRRS(:) = PRRS(:) - ZZW1(:,4)
    PRGS(:) = PRGS(:) + ZZW1(:,3)+ZZW1(:,4)
    PTHS(:) = PTHS(:) + ZZW1(:,4)*(PLSFACT(:)-PLVFACT(:)) ! f(L_f*RRCFRIG)
  END WHERE
!$acc end kernels
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:), 4,'CFRZ_BU_RTH')
  END IF
  IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
    CALL BUDGET (UNPACK(PRRS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),          8,'CFRZ_BU_RRR')
  END IF
  IF (LBUDGET_RI) THEN
!$acc update self(PRIS)
    CALL BUDGET (UNPACK(PRIS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),          9,'CFRZ_BU_RRI')
  END IF
  IF (LBUDGET_RG) THEN
!$acc update self(PRGS)
    CALL BUDGET (UNPACK(PRGS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),         11,'CFRZ_BU_RRG')
  END IF
!
!*       6.2    compute the Dry growth case
!
!$acc kernels
  ZZW1(:,:) = 0.0
  GWORK(:) = PRGT(:)>XRTMIN(6) .AND. PRCT(:)>XRTMIN(2) .AND. PRCS(:)>0.0
  WHERE( GWORK(:) )
#ifndef MNH_BITREP
    ZZW(:) = PLBDAG(:)**(XCXG-XDG-2.0) * PRHODREF(:)**(-XCEXVT)
#else
    ZZW(:) = BR_POW(PLBDAG(:),XCXG-XDG-2.0) * BR_POW(PRHODREF(:),-XCEXVT)
#endif
    ZZW1(:,1) = MIN( PRCS(:),XFCDRYG * PRCT(:) * ZZW(:) )             ! RCDRYG
  END WHERE
  GWORK(:) = PRGT(:)>XRTMIN(6) .AND. PRIT(:)>XRTMIN(4) .AND. PRIS(:)>0.0
  WHERE( GWORK(:) )
#ifndef MNH_BITREP
    ZZW(:) = PLBDAG(:)**(XCXG-XDG-2.0) * PRHODREF(:)**(-XCEXVT)
    ZZW1(:,2) = MIN( PRIS(:),XFIDRYG * EXP( XCOLEXIG*(PZT(:)-XTT) ) &
                                     * PRIT(:) * ZZW(:) )             ! RIDRYG
#else
    ZZW(:) = BR_POW(PLBDAG(:),XCXG-XDG-2.0) * BR_POW(PRHODREF(:),-XCEXVT)
    ZZW1(:,2) = MIN( PRIS(:),XFIDRYG * BR_EXP( XCOLEXIG*(PZT(:)-XTT) ) &
                                     * PRIT(:) * ZZW(:) )             ! RIDRYG
#endif
  END WHERE
!$acc end kernels
!
!*       6.2.1  accretion of aggregates on the graupeln
!
!$acc kernels
  GWORK(:) = PRST(:)>XRTMIN(5) .AND. PRGT(:)>XRTMIN(6) .AND. PRSS(:)>0.0
!$acc end kernels
#ifndef _OPENACC
  IGDRY = COUNTJV( GWORK(:), I1(:) )
#else
  CALL COUNTJV_DEVICE( GWORK(:), I1(:), IGDRY )
#endif

  IF( IGDRY>0 ) THEN
!
!*       6.2.2  allocations
!
    ALLOCATE(ZVECLBDAG(IGDRY))
    ALLOCATE(ZVECLBDAS(IGDRY))
    ALLOCATE(ZVEC1(IGDRY))
    ALLOCATE(ZVEC2(IGDRY))
    ALLOCATE(ZVEC3(IGDRY))
    ALLOCATE(IVEC1(IGDRY))
    ALLOCATE(IVEC2(IGDRY))
!
!*       6.2.3  select the (PLBDAG,PLBDAS) couplet
!
!$acc kernels
    ZVECLBDAG(1:IGDRY) = PLBDAG(I1(1:IGDRY))
    ZVECLBDAS(1:IGDRY) = PLBDAS(I1(1:IGDRY))
!
!*       6.2.4  find the next lower indice for the PLBDAG and for the PLBDAS
!               in the geometrical set of (Lbda_g,Lbda_s) couplet use to
!               tabulate the SDRYG-kernel
!
    ZVEC1(1:IGDRY) = MAX( 1.00001, MIN( REAL(NDRYLBDAG)-0.00001,           &
#ifndef MNH_BITREP
                          XDRYINTP1G * LOG( ZVECLBDAG(1:IGDRY) ) + XDRYINTP2G ) )
#else
                          XDRYINTP1G * BR_LOG( ZVECLBDAG(1:IGDRY) ) + XDRYINTP2G ) )
#endif
    IVEC1(1:IGDRY) = INT( ZVEC1(1:IGDRY) )
    ZVEC1(1:IGDRY) = ZVEC1(1:IGDRY) - REAL( IVEC1(1:IGDRY) )
!
    ZVEC2(1:IGDRY) = MAX( 1.00001, MIN( REAL(NDRYLBDAS)-0.00001,           &
#ifndef MNH_BITREP
                          XDRYINTP1S * LOG( ZVECLBDAS(1:IGDRY) ) + XDRYINTP2S ) )
#else
                          XDRYINTP1S * BR_LOG( ZVECLBDAS(1:IGDRY) ) + XDRYINTP2S ) )
#endif
    IVEC2(1:IGDRY) = INT( ZVEC2(1:IGDRY) )
    ZVEC2(1:IGDRY) = ZVEC2(1:IGDRY) - REAL( IVEC2(1:IGDRY) )
!
!*       6.2.5  perform the bilinear interpolation of the normalized
!               SDRYG-kernel
!
!$acc loop independent
    DO JJ = 1,IGDRY
      ZVEC3(JJ) =  (  XKER_SDRYG(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_SDRYG(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         * ZVEC1(JJ) &
                 - (  XKER_SDRYG(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_SDRYG(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         * (ZVEC1(JJ) - 1.0)
    END DO
!
!$acc loop independent
    DO JJ = 1, IGDRY
      JL = I1(JJ)
#ifndef MNH_BITREP
      ZZW1(JL,3) = MIN( PRSS(JL),XFSDRYG*ZVEC3(JJ)                           & ! RSDRYG
                                      * EXP( XCOLEXSG*(PZT(JL)-XTT) )        &
                    * ZVECLBDAS(JJ)**(XCXS-XBS) * ZVECLBDAG(JJ)**XCXG        &
                    * PRHODREF(JL)**(-XCEXVT-1.)                             &
                         * ( XLBSDRYG1 /   ZVECLBDAG(JJ)**2                  &
                           + XLBSDRYG2 / ( ZVECLBDAG(JJ)   * ZVECLBDAS(JJ) ) &
                           + XLBSDRYG3 /   ZVECLBDAS(JJ)**2 ) )
#else
      ZZW1(JL,3) = MIN( PRSS(JL),XFSDRYG*ZVEC3(JJ)                                & ! RSDRYG
                                      * BR_EXP( XCOLEXSG*(PZT(JL)-XTT) )          &
                    * BR_POW(ZVECLBDAS(JJ),XCXS-XBS) * BR_POW(ZVECLBDAG(JJ),XCXG) &
                    * BR_POW(PRHODREF(JL),-XCEXVT-1.)                             &
                         * ( XLBSDRYG1 /   BR_P2(ZVECLBDAG(JJ))                   &
                           + XLBSDRYG2 / ( ZVECLBDAG(JJ)   * ZVECLBDAS(JJ) )      &
                           + XLBSDRYG3 /   BR_P2(ZVECLBDAS(JJ)) ) )
#endif
    END DO
!$acc end kernels
    DEALLOCATE(ZVECLBDAS)
    DEALLOCATE(ZVECLBDAG)
    DEALLOCATE(IVEC2)
    DEALLOCATE(IVEC1)
    DEALLOCATE(ZVEC3)
    DEALLOCATE(ZVEC2)
    DEALLOCATE(ZVEC1)
  END IF
!
!*       6.2.6  accretion of raindrops on the graupeln
!
!$acc kernels
  GWORK(:) = PRRT(:)>XRTMIN(3) .AND. PRGT(:)>XRTMIN(6) .AND. PRSS(:)>0.0
!$acc end kernels
#ifndef _OPENACC
  IGDRY = COUNTJV( GWORK(:), I1(:) )
#else
  CALL COUNTJV_DEVICE( GWORK(:), I1(:), IGDRY )
#endif
!
  IF( IGDRY>0 ) THEN
!
!*       6.2.7  allocations
!
    ALLOCATE(ZVECLBDAG(IGDRY))
    ALLOCATE(ZVECLBDAR(IGDRY))
    ALLOCATE(ZVEC1(IGDRY))
    ALLOCATE(ZVEC2(IGDRY))
    ALLOCATE(ZVEC3(IGDRY))
    ALLOCATE(IVEC1(IGDRY))
    ALLOCATE(IVEC2(IGDRY))
!
!*       6.2.8  select the (PLBDAG,PLBDAR) couplet
!
!$acc kernels
    ZVECLBDAG(1:IGDRY) = PLBDAG(I1(1:IGDRY))
    ZVECLBDAR(1:IGDRY) = PLBDAR(I1(1:IGDRY))
!
!*       6.2.9  find the next lower indice for the PLBDAG and for the PLBDAR
!               in the geometrical set of (Lbda_g,Lbda_r) couplet use to
!               tabulate the RDRYG-kernel
!
    ZVEC1(1:IGDRY) = MAX( 1.00001, MIN( REAL(NDRYLBDAG)-0.00001,           &
#ifndef MNH_BITREP
                          XDRYINTP1G * LOG( ZVECLBDAG(1:IGDRY) ) + XDRYINTP2G ) )
#else
                          XDRYINTP1G * BR_LOG( ZVECLBDAG(1:IGDRY) ) + XDRYINTP2G ) )
#endif
    IVEC1(1:IGDRY) = INT( ZVEC1(1:IGDRY) )
    ZVEC1(1:IGDRY) = ZVEC1(1:IGDRY) - REAL( IVEC1(1:IGDRY) )
!
    ZVEC2(1:IGDRY) = MAX( 1.00001, MIN( REAL(NDRYLBDAR)-0.00001,           &
#ifndef MNH_BITREP
                          XDRYINTP1R * LOG( ZVECLBDAR(1:IGDRY) ) + XDRYINTP2R ) )
#else
                          XDRYINTP1R * BR_LOG( ZVECLBDAR(1:IGDRY) ) + XDRYINTP2R ) )
#endif
    IVEC2(1:IGDRY) = INT( ZVEC2(1:IGDRY) )
    ZVEC2(1:IGDRY) = ZVEC2(1:IGDRY) - REAL( IVEC2(1:IGDRY) )
!
!*       6.2.10 perform the bilinear interpolation of the normalized
!               RDRYG-kernel
!
!$acc loop independent
    DO JJ = 1,IGDRY
      ZVEC3(JJ) =  (  XKER_RDRYG(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RDRYG(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                                  * ZVEC1(JJ) &
                 - (  XKER_RDRYG(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RDRYG(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                         * (ZVEC1(JJ) - 1.0)
    END DO
!
!$acc loop independent
    DO JJ = 1, IGDRY
      JL = I1(JJ)
#ifndef MNH_BITREP
      ZZW1(JL,4) = MIN( PRRS(JL),XFRDRYG*ZVEC3(JJ)                  & ! RRDRYG
                        * ZVECLBDAR(JJ)**(-4) * ZVECLBDAG(JJ)**XCXG &
                               * PRHODREF(JL)**(-XCEXVT-1.)         &
                    * ( XLBRDRYG1/  ZVECLBDAG(JJ)**2                &
                      + XLBRDRYG2/( ZVECLBDAG(JJ) * ZVECLBDAR(JJ) ) &
                      + XLBRDRYG3/  ZVECLBDAR(JJ)**2 ) )
#else
      ZZW1(JL,4) = MIN( PRRS(JL),XFRDRYG*ZVEC3(JJ)                               & ! RRDRYG
                        * BR_POW(ZVECLBDAR(JJ),-4.) * BR_POW(ZVECLBDAG(JJ),XCXG) &
                               * BR_POW(PRHODREF(JL),-XCEXVT-1.)                 &
                    * ( XLBRDRYG1/  BR_P2(ZVECLBDAG(JJ))                         &
                      + XLBRDRYG2/( ZVECLBDAG(JJ) * ZVECLBDAR(JJ) )              &
                      + XLBRDRYG3/  BR_P2(ZVECLBDAR(JJ)) ) )
#endif
    END DO
!$acc end kernels
    DEALLOCATE(ZVECLBDAR)
    DEALLOCATE(ZVECLBDAG)
    DEALLOCATE(IVEC2)
    DEALLOCATE(IVEC1)
    DEALLOCATE(ZVEC3)
    DEALLOCATE(ZVEC2)
    DEALLOCATE(ZVEC1)
  END IF
!
!$acc kernels
  PRDRYG(:) = ZZW1(:,1) + ZZW1(:,2) + ZZW1(:,3) + ZZW1(:,4)
!
!*       6.3    compute the Wet growth case
!
  PRWETG(:) = 0.0
  GWORK(:) = PRGT(:)>XRTMIN(6)
  WHERE( GWORK(:) )
#ifndef MNH_BITREP
    ZZW1(:,5) = MIN( PRIS(:),                                    &
                ZZW1(:,2) / (XCOLIG*EXP(XCOLEXIG*(PZT(:)-XTT)) ) ) ! RIWETG
    ZZW1(:,6) = MIN( PRSS(:),                                    &
                ZZW1(:,3) / (XCOLSG*EXP(XCOLEXSG*(PZT(:)-XTT)) ) ) ! RSWETG
#else
    ZZW1(:,5) = MIN( PRIS(:),                                    &
                ZZW1(:,2) / (XCOLIG*BR_EXP(XCOLEXIG*(PZT(:)-XTT)) ) ) ! RIWETG
    ZZW1(:,6) = MIN( PRSS(:),                                    &
                ZZW1(:,3) / (XCOLSG*BR_EXP(XCOLEXSG*(PZT(:)-XTT)) ) ) ! RSWETG
#endif
!
    ZZW(:) = PRVT(:)*PPRES(:)/((XMV/XMD)+PRVT(:)) ! Vapor pressure
    ZZW(:) =   PKA(:)*(XTT-PZT(:)) +                              &
             ( PDV(:)*(XLVTT + ( XCPV - XCL ) * ( PZT(:) - XTT )) &
                           *(XESTT-ZZW(:))/(XRV*PZT(:))           )
!
! compute RWETG
!
    PRWETG(:)=MAX( 0.0,                                               &
#ifndef MNH_BITREP
                 ( ZZW(:) * ( X0DEPG*       PLBDAG(:)**XEX0DEPG +     &
                              X1DEPG*PCJ(:)*PLBDAG(:)**XEX1DEPG ) +   &
#else
                 ( ZZW(:) * ( X0DEPG*       BR_POW(PLBDAG(:),XEX0DEPG) +     &
                              X1DEPG*PCJ(:)*BR_POW(PLBDAG(:),XEX1DEPG) ) +   &
#endif
                 ( ZZW1(:,5)+ZZW1(:,6) ) *                            &
                 ( PRHODREF(:)*(XLMTT+(XCI-XCL)*(XTT-PZT(:)))   ) ) / &
                            ( PRHODREF(:)*(XLMTT-XCL*(XTT-PZT(:))) )   )
  END WHERE
!
!*       6.4    Select Wet or Dry case
!
  IF     ( KRR == 7 ) THEN
   GWORK(:) = PRGT(:)>XRTMIN(6) .AND. PZT(:)<XTT .and. PRDRYG(:)>=PRWETG(:) .AND. PRWETG(:)>0.0 ! Wet case
   WHERE( GWORK(:) )
     ZZW(:) = PRWETG(:) - ZZW1(:,5) - ZZW1(:,6) ! RCWETG+RRWETG
!
! limitation of the available rainwater mixing ratio (RRWETH < RRS !)
!
    ZZW1(:,7) = MAX( 0.0,MIN( ZZW(:),PRRS(:)+ZZW1(:,1) ) )
    PUSW(:)   = ZZW1(:,7) / ZZW(:)
    ZZW1(:,5) = ZZW1(:,5)*PUSW(:)
    ZZW1(:,6) = ZZW1(:,6)*PUSW(:)
    PRWETG(:) = ZZW1(:,7) + ZZW1(:,5) + ZZW1(:,6)
!
    PRCS(:) = PRCS(:) - ZZW1(:,1)
    PRIS(:) = PRIS(:) - ZZW1(:,5)
    PRSS(:) = PRSS(:) - ZZW1(:,6)
!
! assume a linear percent of conversion of graupel into hail
!
    PRGS(:) = PRGS(:) + PRWETG(:)                     !     Wet growth
    ZZW(:)  = PRGS(:)*PRDRYG(:)/(PRWETG(:)+PRDRYG(:)) !        and
    PRGS(:) = PRGS(:) - ZZW(:)                        !   partial conversion
    PRHS(:) = PRHS(:) + ZZW(:)                        ! of the graupel into hail
!
    PRRS(:) = MAX( 0.0,PRRS(:) - ZZW1(:,7) + ZZW1(:,1) )
    PTHS(:) = PTHS(:) + ZZW1(:,7)*(PLSFACT(:)-PLVFACT(:))
                                                 ! f(L_f*(RCWETG+RRWETG))
   END WHERE
   ELSE IF( KRR == 6 ) THEN
    GWORK(:) = PRGT(:)>XRTMIN(6) .AND. PZT(:)<XTT .AND. PRDRYG(:)>=PRWETG(:) .AND. PRWETG(:)>0.0 ! Wet case
    WHERE( GWORK(:) )
    PRCS(:) = PRCS(:) - ZZW1(:,1)
    PRIS(:) = PRIS(:) - ZZW1(:,5)
    PRSS(:) = PRSS(:) - ZZW1(:,6)
    PRGS(:) = PRGS(:) + PRWETG(:)
!
    PRRS(:) = PRRS(:) - PRWETG(:) + ZZW1(:,5) + ZZW1(:,6) + ZZW1(:,1)
    PTHS(:) = PTHS(:) + (PRWETG(:)-ZZW1(:,5)-ZZW1(:,6))*(PLSFACT(:)-PLVFACT(:))
                                                 ! f(L_f*(RCWETG+RRWETG))
   END WHERE
 END IF
!$acc end kernels
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'WETG_BU_RTH')
  END IF
  IF (LBUDGET_RC) THEN
!$acc update self(PRCS)
    CALL BUDGET (UNPACK(PRCS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),7,'WETG_BU_RRC')
  END IF
  IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
    CALL BUDGET (UNPACK(PRRS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),8,'WETG_BU_RRR')
  END IF
  IF (LBUDGET_RI) THEN
!$acc update self(PRIS)
    CALL BUDGET (UNPACK(PRIS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),9,'WETG_BU_RRI')
  END IF
  IF (LBUDGET_RS) THEN
!$acc update self(PRSS)
    CALL BUDGET (UNPACK(PRSS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),10,'WETG_BU_RRS')
  END IF
  IF (LBUDGET_RG) THEN
!$acc update self(PRGS)
    CALL BUDGET (UNPACK(PRGS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),11,'WETG_BU_RRG')
  END IF
  IF ( KRR == 7 ) THEN
    IF (LBUDGET_RH) THEN
!$acc update self(PRHS)
      CALL BUDGET (UNPACK(PRHS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),12,'WETG_BU_RRH')
    END IF
  END IF

!
!$acc kernels
  GWORK(:) = PRGT(:)>XRTMIN(6) .AND. PZT(:)<XTT .AND. PRDRYG(:)<PRWETG(:) .AND. PRDRYG(:)>0.0 ! Dry case
  WHERE( GWORK(:) )
    PRCS(:) = PRCS(:) - ZZW1(:,1)
    PRIS(:) = PRIS(:) - ZZW1(:,2)
    PRSS(:) = PRSS(:) - ZZW1(:,3)
    PRRS(:) = PRRS(:) - ZZW1(:,4)
    PRGS(:) = PRGS(:) + PRDRYG(:)
    PTHS(:) = PTHS(:) + (ZZW1(:,1)+ZZW1(:,4))*(PLSFACT(:)-PLVFACT(:)) !
                      ! f(L_f*(RCDRYG+RRDRYG))
  END WHERE
!$acc end kernels
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'DRYG_BU_RTH')
  END IF
  IF (LBUDGET_RC) THEN
!$acc update self(PRCS)
    CALL BUDGET (UNPACK(PRCS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),7,'DRYG_BU_RRC')
  END IF
  IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
    CALL BUDGET (UNPACK(PRRS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),8,'DRYG_BU_RRR')
  END IF
  IF (LBUDGET_RI) THEN
!$acc update self(PRIS)
    CALL BUDGET (UNPACK(PRIS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),9,'DRYG_BU_RRI')
  END IF
  IF (LBUDGET_RS) THEN
!$acc update self(PRSS)
    CALL BUDGET (UNPACK(PRSS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),10,'DRYG_BU_RRS')
  END IF
  IF (LBUDGET_RG) THEN
!$acc update self(PRGS)
    CALL BUDGET (UNPACK(PRGS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),11,'DRYG_BU_RRG')
  END IF
!
!      WHERE ( PZT(:) > XTT ) ! RSWETG case only
!        PRSS(:) = PRSS(:) - ZZW1(:,6)
!        PRGS(:) = PRGS(:) + ZZW1(:,6)
!      END WHERE
!
!*       6.5    Melting of the graupeln
!
!$acc kernels
   GWORK(:) = PRGT(:)>XRTMIN(6) .AND. PRGS(:)>0.0 .AND. PZT(:)>XTT
   WHERE( GWORK(:) )
    ZZW(:) = PRVT(:)*PPRES(:)/((XMV/XMD)+PRVT(:)) ! Vapor pressure
    ZZW(:) =  PKA(:)*(XTT-PZT(:)) +                                 &
               ( PDV(:)*(XLVTT + ( XCPV - XCL ) * ( PZT(:) - XTT )) &
                           *(XESTT-ZZW(:))/(XRV*PZT(:))             )
!
! compute RGMLTR
!
    ZZW(:)  = MIN( PRGS(:), MAX( 0.0,( -ZZW(:) *                     &
#ifndef MNH_BITREP
                           ( X0DEPG*       PLBDAG(:)**XEX0DEPG +     &
                             X1DEPG*PCJ(:)*PLBDAG(:)**XEX1DEPG ) -   &
#else
                           ( X0DEPG*       BR_POW(PLBDAG(:),XEX0DEPG) +     &
                             X1DEPG*PCJ(:)*BR_POW(PLBDAG(:),XEX1DEPG) ) -   &
#endif
                                     ( ZZW1(:,1)+ZZW1(:,4) ) *       &
                              ( PRHODREF(:)*XCL*(XTT-PZT(:))) ) /    &
                                             ( PRHODREF(:)*XLMTT ) ) )
    PRRS(:) = PRRS(:) + ZZW(:)
    PRGS(:) = PRGS(:) - ZZW(:)
    PTHS(:) = PTHS(:) - ZZW(:)*(PLSFACT(:)-PLVFACT(:)) ! f(L_f*(-RGMLTR))
  END WHERE
!$acc end kernels
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
      CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'GMLT_BU_RTH')
  END IF
  IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
    CALL BUDGET (UNPACK(PRRS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),8,'GMLT_BU_RRR')
  END IF
  IF (LBUDGET_RG) THEN
!$acc update self(PRGS)
    CALL BUDGET (UNPACK(PRGS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),11,'GMLT_BU_RRG')
  END IF
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRCS,"RAIN_ICE_FAST_RG end:PRCS")
  CALL MPPDB_CHECK(PRRS,"RAIN_ICE_FAST_RG end:PRRS")
  CALL MPPDB_CHECK(PRIS,"RAIN_ICE_FAST_RG end:PRIS")
  CALL MPPDB_CHECK(PRSS,"RAIN_ICE_FAST_RG end:PRSS")
  CALL MPPDB_CHECK(PRGS,"RAIN_ICE_FAST_RG end:PRGS")
  CALL MPPDB_CHECK(PRHS,"RAIN_ICE_FAST_RG end:PRHS")
  CALL MPPDB_CHECK(PTHS,"RAIN_ICE_FAST_RG end:PTHS")
  CALL MPPDB_CHECK(PUSW,"RAIN_ICE_FAST_RG end:PUSW")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRDRYG,"RAIN_ICE_FAST_RG end:PRDRYG")
  CALL MPPDB_CHECK(PRWETG,"RAIN_ICE_FAST_RG end:PRWETG")
END IF

END SUBROUTINE RAIN_ICE_FAST_RG

END MODULE MODE_RAIN_ICE_FAST_RG
