!MNH_LIC Copyright 1995-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
! Modifications:
!  P. Wautelet 25/02/2019: split rain_ice (cleaner and easier to maintain/debug)
!  P. Wautelet 26/04/2019: replace non-standard FLOAT function by REAL function
!  P. Wautelet 03/06/2019: remove PACK/UNPACK intrinsics (to get more performance and better OpenACC support)
!  P. Wautelet 05/06/2019: optimisations
!-----------------------------------------------------------------
MODULE MODE_RAIN_ICE_FAST_RS

  IMPLICIT NONE

  PRIVATE

  PUBLIC :: RAIN_ICE_FAST_RS

CONTAINS

SUBROUTINE RAIN_ICE_FAST_RS(PTSTEP, OMICRO, PRHODREF, PRVT, PRCT, PRRT, PRST, PRHODJ, PPRES, PZT, &
                            PLBDAR, PLBDAS, PLSFACT, PLVFACT, PCJ, PKA, PDV, PRHODJ3D, PTHS3D, &
                            PRCS, PRRS, PRSS, PRGS, PTHS)
!
!*      0. DECLARATIONS
!          ------------
!
use MODD_BUDGET,         only: LBUDGET_RC, LBUDGET_RG, LBUDGET_RR, LBUDGET_RS, LBUDGET_TH
use MODD_CST,            only: XCL, XCPV, XESTT, XLMTT, XLVTT, XMD, XMV, XRV, XTT
use MODD_RAIN_ICE_DESCR, only: XBS, XCEXVT, XCXS, XRTMIN
use MODD_RAIN_ICE_PARAM, only: NACCLBDAR, NACCLBDAS, NGAMINC, X0DEPS, X1DEPS, XACCINTP1R, XACCINTP1S, XACCINTP2R, XACCINTP2S, &
                               XCRIMSG, XCRIMSS, XEX0DEPS, XEX1DEPS, XEXCRIMSG, XEXCRIMSS, XEXSRIMCG, XFRACCSS,               &
                               XFSACCRG, XFSCVMG, XGAMINC_RIM1, XGAMINC_RIM1, XGAMINC_RIM2, XKER_RACCS,                       &
                               XKER_RACCSS, XKER_SACCRG, XLBRACCS1, XLBRACCS2, XLBRACCS3, XLBSACCR1, XLBSACCR2, XLBSACCR3,    &
                               XRIMINTP1, XRIMINTP2, XSRIMCG
!
use mode_mppdb
#ifndef _OPENACC
use mode_tools,                        only: Countjv
#else
use mode_tools,                        only: Countjv_device
#endif
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
use MODI_BUDGET
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
REAL,                       intent(in)    :: PTSTEP   ! Double Time step
                                                      ! (single if cold start)
LOGICAL,  DIMENSION(:,:,:), intent(in)    :: OMICRO   ! Test where to compute all processes
REAL,     DIMENSION(:),     intent(in)    :: PRHODREF ! RHO Dry REFerence
REAL,     DIMENSION(:),     intent(in)    :: PRVT     ! Water vapor m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRCT     ! Cloud water m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRRT     ! Rain water m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRST     ! Snow/aggregate m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRHODJ   ! RHO times Jacobian
REAL,     DIMENSION(:),     intent(in)    :: PPRES    ! Pressure
REAL,     DIMENSION(:),     intent(in)    :: PZT      ! Temperature
REAL,     DIMENSION(:),     intent(in)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL,     DIMENSION(:),     intent(in)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL,     DIMENSION(:),     intent(in)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL,     DIMENSION(:),     intent(in)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL,     DIMENSION(:),     intent(in)    :: PCJ      ! Function to compute the ventilation coefficient
REAL,     DIMENSION(:),     intent(in)    :: PKA      ! Thermal conductivity of the air
REAL,     DIMENSION(:),     intent(in)    :: PDV      ! Diffusivity of water vapor in the air
REAL,     DIMENSION(:,:,:), INTENT(IN)    :: PRHODJ3D ! Dry density * Jacobian
REAL,     DIMENSION(:,:,:), INTENT(IN)    :: PTHS3D   ! Theta source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRCS     ! Cloud water m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRRS     ! Rain water m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRSS     ! Snow/aggregate m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRGS     ! Graupel m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PTHS     ! Theta source
!
! IN variables
!
!$acc declare present(OMICRO, PRHODREF, PRVT, PRCT, PRRT, PRST, PRHODJ, &
!$acc &               PPRES, PZT, PLBDAR, PLBDAS, PLSFACT, PLVFACT,     &
!$acc &               PCJ, PKA, PDV, PRHODJ3D, PTHS3D)                  &
!
! INOUT variables
!
!$acc &       present(PRCS, PRRS, PRSS, PRGS, PTHS)
!
! OUT variables
!
!NONE
!
!*       0.2  declaration of local variables
!
INTEGER                            :: IGRIM, IGACC
INTEGER                            :: JJ, JL
INTEGER, DIMENSION(:), ALLOCATABLE :: I1
INTEGER, DIMENSION(:), ALLOCATABLE :: IVEC1, IVEC2      ! Vectors of indices for interpolations
LOGICAL, DIMENSION(:), ALLOCATABLE :: GWORK
REAL,    DIMENSION(:), ALLOCATABLE :: ZZW               ! Work array
REAL,    DIMENSION(:), ALLOCATABLE :: ZVEC1,ZVEC2,ZVEC3 ! Work vectors for interpolations
REAL,    DIMENSION(:), ALLOCATABLE :: ZVECLBDAR, ZVECLBDAS
REAL,    DIMENSION(:), ALLOCATABLE :: ZZW1, ZZW2, ZZW3, ZZW4 ! Work arrays
!
!$acc declare device_resident(I1, IVEC1, IVEC2, GWORK, ZZW, ZVEC1, ZVEC2, ZVEC3, &
!$acc &                       ZVECLBDAR, ZVECLBDAS, ZZW1, ZZW2, ZZW3, ZZW4)
!
!$acc declare copyin(XKER_RACCS, XKER_RACCSS, XKER_SACCRG)
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(OMICRO,"RAIN_ICE_FAST_RS beg:OMICRO")
  CALL MPPDB_CHECK(PRHODREF,"RAIN_ICE_FAST_RS beg:PRHODREF")
  CALL MPPDB_CHECK(PRVT,"RAIN_ICE_FAST_RS beg:PRVT")
  CALL MPPDB_CHECK(PRCT,"RAIN_ICE_FAST_RS beg:PRCT")
  CALL MPPDB_CHECK(PRRT,"RAIN_ICE_FAST_RS beg:PRRT")
  CALL MPPDB_CHECK(PRST,"RAIN_ICE_FAST_RS beg:PRST")
  CALL MPPDB_CHECK(PRHODJ,"RAIN_ICE_FAST_RS beg:PRHODJ")
  CALL MPPDB_CHECK(PPRES,"RAIN_ICE_FAST_RS beg:PPRES")
  CALL MPPDB_CHECK(PZT,"RAIN_ICE_FAST_RS beg:PZT")
  CALL MPPDB_CHECK(PLBDAR,"RAIN_ICE_FAST_RS beg:PLBDAR")
  CALL MPPDB_CHECK(PLBDAS,"RAIN_ICE_FAST_RS beg:PLBDAS")
  CALL MPPDB_CHECK(PLSFACT,"RAIN_ICE_FAST_RS beg:PLSFACT")
  CALL MPPDB_CHECK(PLVFACT,"RAIN_ICE_FAST_RS beg:PLVFACT")
  CALL MPPDB_CHECK(PCJ,"RAIN_ICE_FAST_RS beg:PCJ")
  CALL MPPDB_CHECK(PKA,"RAIN_ICE_FAST_RS beg:PKA")
  CALL MPPDB_CHECK(PDV,"RAIN_ICE_FAST_RS beg:PDV")
  CALL MPPDB_CHECK(PRHODJ3D,"RAIN_ICE_FAST_RS beg:PRHODJ3D")
  CALL MPPDB_CHECK(PTHS3D,"RAIN_ICE_FAST_RS beg:PTHS3D")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRCS,"RAIN_ICE_FAST_RS beg:PRCS")
  CALL MPPDB_CHECK(PRRS,"RAIN_ICE_FAST_RS beg:PRRS")
  CALL MPPDB_CHECK(PRSS,"RAIN_ICE_FAST_RS beg:PRSS")
  CALL MPPDB_CHECK(PRGS,"RAIN_ICE_FAST_RS beg:PRGS")
  CALL MPPDB_CHECK(PTHS,"RAIN_ICE_FAST_RS beg:PTHS")
END IF
!
ALLOCATE( I1   (size(PRHODREF)) )
ALLOCATE( GWORK(size(PRHODREF)) )
ALLOCATE( ZZW  (size(PRHODREF)) )
!
!*       5.1    cloud droplet riming of the aggregates
!
!$acc kernels
GWORK(:) = PRCT(:)>XRTMIN(2) .AND. PRST(:)>XRTMIN(5) .AND. PRCS(:)>0.0 .AND. PZT(:)<XTT
!$acc end kernels
#ifndef _OPENACC
IGRIM = COUNTJV( GWORK(:), I1(:) )
#else
CALL COUNTJV_DEVICE( GWORK(:), I1(:), IGRIM )
#endif
  !
  IF( IGRIM>0 ) THEN
!
!        5.1.0  allocations
!
    ALLOCATE(ZVECLBDAS(IGRIM))
    ALLOCATE(ZVEC1(IGRIM))
    ALLOCATE(ZVEC2(IGRIM))
    ALLOCATE(IVEC2(IGRIM))
    ALLOCATE(ZZW1(IGRIM))
    ALLOCATE(ZZW2(IGRIM))
    ALLOCATE(ZZW3(IGRIM))
!
!        5.1.1  select the PLBDAS
!
!$acc kernels
    ZVECLBDAS(1:IGRIM) = PLBDAS(I1(1:IGRIM))
!
!        5.1.2  find the next lower indice for the PLBDAS in the geometrical
!               set of Lbda_s used to tabulate some moments of the incomplete
!               gamma function
!
    ZVEC2(1:IGRIM) = MAX( 1.00001, MIN( REAL(NGAMINC)-0.00001,           &
#ifndef MNH_BITREP
                          XRIMINTP1 * LOG( ZVECLBDAS(1:IGRIM) ) + XRIMINTP2 ) )
#else
                          XRIMINTP1 * BR_LOG( ZVECLBDAS(1:IGRIM) ) + XRIMINTP2 ) )
#endif
    IVEC2(1:IGRIM) = INT( ZVEC2(1:IGRIM) )
    ZVEC2(1:IGRIM) = ZVEC2(1:IGRIM) - REAL( IVEC2(1:IGRIM) )
!
!        5.1.3  perform the linear interpolation of the normalized
!               "2+XDS"-moment of the incomplete gamma function
!
    ZVEC1(1:IGRIM) =   XGAMINC_RIM1( IVEC2(1:IGRIM)+1 )* ZVEC2(1:IGRIM)      &
                     - XGAMINC_RIM1( IVEC2(1:IGRIM)   )*(ZVEC2(1:IGRIM) - 1.0)
!
!        5.1.4  riming of the small sized aggregates
!
!$acc loop independent
    DO JJ = 1, IGRIM
      JL = I1(JJ)
#ifndef MNH_BITREP
      ZZW1(JJ) = MIN( PRCS(JL),                           &
                     XCRIMSS * ZVEC1(JJ) * PRCT(JL)       & ! RCRIMSS
                             *   ZVECLBDAS(JJ)**XEXCRIMSS &
                             * PRHODREF(JL)**(-XCEXVT) )
#else
      ZZW1(JJ) = MIN( PRCS(JL),                                &
                     XCRIMSS * ZVEC1(JJ) * PRCT(JL)            & ! RCRIMSS
                             * BR_POW(ZVECLBDAS(JJ),XEXCRIMSS) &
                             * BR_POW(PRHODREF(JL),-XCEXVT) )
#endif
      PRCS(JL) = PRCS(JL) - ZZW1(JJ)
      PRSS(JL) = PRSS(JL) + ZZW1(JJ)
      PTHS(JL) = PTHS(JL) + ZZW1(JJ)*(PLSFACT(JL)-PLVFACT(JL)) ! f(L_f*(RCRIMSS))
    END DO
!
!        5.1.5  perform the linear interpolation of the normalized
!               "XBS"-moment of the incomplete gamma function
!
    ZVEC1(1:IGRIM) =  XGAMINC_RIM2( IVEC2(1:IGRIM)+1 )* ZVEC2(1:IGRIM)      &
                    - XGAMINC_RIM2( IVEC2(1:IGRIM)   )*(ZVEC2(1:IGRIM) - 1.0)
!
!        5.1.6  riming-conversion of the large sized aggregates into graupeln
!
!
!$acc loop independent
    DO JJ = 1, IGRIM
      JL = I1(JJ)
      IF ( PRSS(JL) > 0.0 ) THEN
#ifndef MNH_BITREP
        ZZW2(JJ) = MIN( PRCS(JL),                       &
                    XCRIMSG * PRCT(JL)                  & ! RCRIMSG
                            * ZVECLBDAS(JJ)**XEXCRIMSG  &
                            * PRHODREF(JL)**(-XCEXVT)   &
                            - ZZW1(JJ) )
        ZZW3(JJ) = MIN( PRSS(JL),                                          &
                        XSRIMCG * ZVECLBDAS(JJ)**XEXSRIMCG                 & ! RSRIMCG
                                * (1.0 - ZVEC1(JJ) )/(PTSTEP*PRHODREF(JL)) )
#else
        ZZW2(JJ) = MIN( PRCS(JL),                             &
                    XCRIMSG * PRCT(JL)                        & ! RCRIMSG
                            * BR_POW(ZVECLBDAS(JJ),XEXCRIMSG) &
                            * BR_POW(PRHODREF(JL),-XCEXVT)    &
                            - ZZW1(JJ) )
        ZZW3(JJ) = MIN( PRSS(JL),                                          &
                        XSRIMCG * BR_POW(ZVECLBDAS(JJ),XEXSRIMCG)          & ! RSRIMCG
                                * (1.0 - ZVEC1(JJ) )/(PTSTEP*PRHODREF(JL)) )
#endif
        PRCS(JL) = PRCS(JL) - ZZW2(JJ)
        PRSS(JL) = PRSS(JL) - ZZW3(JJ)
        PRGS(JL) = PRGS(JL) + ZZW2(JJ)+ZZW3(JJ)
        PTHS(JL) = PTHS(JL) + ZZW2(JJ)*(PLSFACT(JL)-PLVFACT(JL)) ! f(L_f*(RCRIMSG))
      END IF
    END DO
!$acc end kernels
    DEALLOCATE(ZZW3)
    DEALLOCATE(ZZW2)
    DEALLOCATE(ZZW1)
    DEALLOCATE(IVEC2)
    DEALLOCATE(ZVEC2)
    DEALLOCATE(ZVEC1)
    DEALLOCATE(ZVECLBDAS)
  END IF
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'RIM_BU_RTH')
  END IF
  IF (LBUDGET_RC) THEN
!$acc update self(PRCS)
    CALL BUDGET (UNPACK(PRCS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),7,'RIM_BU_RRC')
  END IF
  IF (LBUDGET_RS) THEN
!$acc update self(PRSS)
    CALL BUDGET (UNPACK(PRSS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),10,'RIM_BU_RRS')
  END IF
  IF (LBUDGET_RG) THEN
!$acc update self(PRGS)
    CALL BUDGET (UNPACK(PRGS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),11,'RIM_BU_RRG')
  END IF
!
!*       5.2    rain accretion onto the aggregates
!
!$acc kernels
  GWORK(:) = PRRT(:)>XRTMIN(3) .AND. PRST(:)>XRTMIN(5) .AND. PRRS(:)>0.0 .AND. PZT(:)<XTT
!$acc end kernels
#ifndef _OPENACC
  IGACC = COUNTJV( GWORK(:), I1(:) )
#else
  CALL COUNTJV_DEVICE( GWORK(:), I1(:), IGACC )
#endif
  !
  IF( IGACC>0 ) THEN
!
!        5.2.0  allocations
!
    ALLOCATE(ZVECLBDAR(IGACC))
    ALLOCATE(ZVECLBDAS(IGACC))
    ALLOCATE(ZVEC1(IGACC))
    ALLOCATE(ZVEC2(IGACC))
    ALLOCATE(ZVEC3(IGACC))
    ALLOCATE(IVEC1(IGACC))
    ALLOCATE(IVEC2(IGACC))
    ALLOCATE(ZZW2(IGACC))
    ALLOCATE(ZZW3(IGACC))
    ALLOCATE(ZZW4(IGACC))
!
!        5.2.1  select the (PLBDAS,PLBDAR) couplet
!
!$acc kernels
    ZVECLBDAS(1:IGACC) = PLBDAS(I1(1:IGACC))
    ZVECLBDAR(1:IGACC) = PLBDAR(I1(1:IGACC))
!
!        5.2.2  find the next lower indice for the PLBDAS and for the PLBDAR
!               in the geometrical set of (Lbda_s,Lbda_r) couplet use to
!               tabulate the RACCSS-kernel
!
    ZVEC1(1:IGACC) = MAX( 1.00001, MIN( REAL(NACCLBDAS)-0.00001,           &
#ifndef MNH_BITREP
                          XACCINTP1S * LOG( ZVECLBDAS(1:IGACC) ) + XACCINTP2S ) )
#else
                          XACCINTP1S * BR_LOG( ZVECLBDAS(1:IGACC) ) + XACCINTP2S ) )
#endif
    IVEC1(1:IGACC) = INT( ZVEC1(1:IGACC) )
    ZVEC1(1:IGACC) = ZVEC1(1:IGACC) - REAL( IVEC1(1:IGACC) )
!
    ZVEC2(1:IGACC) = MAX( 1.00001, MIN( REAL(NACCLBDAR)-0.00001,           &
#ifndef MNH_BITREP
                          XACCINTP1R * LOG( ZVECLBDAR(1:IGACC) ) + XACCINTP2R ) )
#else
                          XACCINTP1R * BR_LOG( ZVECLBDAR(1:IGACC) ) + XACCINTP2R ) )
#endif
    IVEC2(1:IGACC) = INT( ZVEC2(1:IGACC) )
    ZVEC2(1:IGACC) = ZVEC2(1:IGACC) - REAL( IVEC2(1:IGACC) )
!
!        5.2.3  perform the bilinear interpolation of the normalized
!               RACCSS-kernel
!
!$acc loop independent
    DO JJ = 1,IGACC
      ZVEC3(JJ) =  (  XKER_RACCSS(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RACCSS(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                          * ZVEC1(JJ) &
                 - (  XKER_RACCSS(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RACCSS(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                          * (ZVEC1(JJ) - 1.0)
    END DO
!
!        5.2.4  raindrop accretion on the small sized aggregates
!
!$acc loop independent
    DO JJ = 1, IGACC
      JL = I1(JJ)
#ifndef MNH_BITREP
      ZZW2(JJ) =                                                          & !! coef of RRACCS
              XFRACCSS * ZVECLBDAS(JJ)**XCXS * PRHODREF(JL)**(-XCEXVT-1.) &
         *( XLBRACCS1 / ZVECLBDAS(JJ)**2                                  &
          + XLBRACCS2 / ( ZVECLBDAS(JJ)    * ZVECLBDAR(JJ)    )           &
          + XLBRACCS3 / ZVECLBDAR(JJ)**2 ) / ZVECLBDAR(JJ)**4
#else
      ZZW2(JJ) =                                                                      & !! coef of RRACCS
              XFRACCSS * BR_POW(ZVECLBDAS(JJ),XCXS) * BR_POW(PRHODREF(JL),-XCEXVT-1.) &
         *( XLBRACCS1 / BR_P2(ZVECLBDAS(JJ))                                          &
          + XLBRACCS2 / ( ZVECLBDAS(JJ)    * ZVECLBDAR(JJ)    )                       &
          + XLBRACCS3 / BR_P2(ZVECLBDAR(JJ)) ) / BR_POW(ZVECLBDAR(JJ),4.0)
#endif
      ZZW4(JJ) = MIN( PRRS(JL),ZZW2(JJ)*ZVEC3(JJ) )           ! RRACCSS
      PRRS(JL) = PRRS(JL) - ZZW4(JJ)
      PRSS(JL) = PRSS(JL) + ZZW4(JJ)
      PTHS(JL) = PTHS(JL) + ZZW4(JJ)*(PLSFACT(JL)-PLVFACT(JL)) ! f(L_f*(RRACCSS))
    END DO
!
!        5.2.4b perform the bilinear interpolation of the normalized
!               RACCS-kernel
!
!$acc loop independent
    DO JJ = 1,IGACC
      ZVEC3(JJ) =  (   XKER_RACCS(IVEC2(JJ)+1,IVEC1(JJ)+1)* ZVEC1(JJ)          &
                    -  XKER_RACCS(IVEC2(JJ)+1,IVEC1(JJ)  )*(ZVEC1(JJ) - 1.0) ) &
                                                                   * ZVEC2(JJ) &
                 - (   XKER_RACCS(IVEC2(JJ)  ,IVEC1(JJ)+1)* ZVEC1(JJ)          &
                    -  XKER_RACCS(IVEC2(JJ)  ,IVEC1(JJ)  )*(ZVEC1(JJ) - 1.0) ) &
                                                           * (ZVEC2(JJ) - 1.0)
    END DO
!$acc loop independent
    DO JJ = 1, IGACC
      ZZW2(JJ) = ZZW2(JJ) * ZVEC3(JJ)
    END DO
                                                                       !! RRACCS!
!        5.2.5  perform the bilinear interpolation of the normalized
!               SACCRG-kernel
!
!$acc loop independent
    DO JJ = 1,IGACC
      ZVEC3(JJ) =  (  XKER_SACCRG(IVEC2(JJ)+1,IVEC1(JJ)+1)* ZVEC1(JJ)          &
                    - XKER_SACCRG(IVEC2(JJ)+1,IVEC1(JJ)  )*(ZVEC1(JJ) - 1.0) ) &
                                                          * ZVEC2(JJ) &
                 - (  XKER_SACCRG(IVEC2(JJ)  ,IVEC1(JJ)+1)* ZVEC1(JJ)          &
                    - XKER_SACCRG(IVEC2(JJ)  ,IVEC1(JJ)  )*(ZVEC1(JJ) - 1.0) ) &
                                                          * (ZVEC2(JJ) - 1.0)
    END DO
!
!        5.2.6  raindrop accretion-conversion of the large sized aggregates
!               into graupeln
!
!$acc loop independent
    DO JJ = 1, IGACC
      JL = I1(JJ)
      IF ( PRSS(JL) > 0.0 ) THEN
        ZZW2(JJ) = MAX( MIN( PRRS(JL),ZZW2(JJ)-ZZW4(JJ) ),0.0 )       ! RRACCSG
        IF ( ZZW2(JJ) > 0.0 ) THEN
#ifndef MNH_BITREP
          ZZW3(JJ) = MIN( PRSS(JL),XFSACCRG*ZVEC3(JJ)*                  & ! RSACCRG
                 ZVECLBDAS(JJ)**(XCXS-XBS) * PRHODREF(JL)**(-XCEXVT-1.) &
             * ( XLBSACCR1 / ZVECLBDAR(JJ)**2                           &
               + XLBSACCR2 /( ZVECLBDAR(JJ) * ZVECLBDAS(JJ) )           &
               + XLBSACCR3 / ZVECLBDAS(JJ)**2 ) / ZVECLBDAR(JJ) )
#else
          ZZW3(JJ) = MIN( PRSS(JL),XFSACCRG*ZVEC3(JJ)*                  & ! RSACCRG
                 BR_POW(ZVECLBDAS(JJ),XCXS-XBS) * BR_POW(PRHODREF(JL),-XCEXVT-1.) &
             * ( XLBSACCR1 / BR_P2(ZVECLBDAR(JJ))                           &
               + XLBSACCR2 /( ZVECLBDAR(JJ) * ZVECLBDAS(JJ) )           &
               + XLBSACCR3 / BR_P2(ZVECLBDAS(JJ)) ) / ZVECLBDAR(JJ) )
#endif
          PRRS(JL) = PRRS(JL) - ZZW2(JJ)
          PRSS(JL) = PRSS(JL) - ZZW3(JJ)
          PRGS(JL) = PRGS(JL) + ZZW2(JJ)+ZZW3(JJ)
          PTHS(JL) = PTHS(JL) + ZZW2(JJ)*(PLSFACT(JL)-PLVFACT(JL)) !
                                ! f(L_f*(RRACCSG))
        END IF
      END IF
    END DO
!$acc end kernels
    DEALLOCATE(ZZW4)
    DEALLOCATE(ZZW3)
    DEALLOCATE(ZZW2)
    DEALLOCATE(IVEC2)
    DEALLOCATE(IVEC1)
    DEALLOCATE(ZVEC3)
    DEALLOCATE(ZVEC2)
    DEALLOCATE(ZVEC1)
    DEALLOCATE(ZVECLBDAS)
    DEALLOCATE(ZVECLBDAR)
  END IF
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'ACC_BU_RTH')
  END IF
  IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
    CALL BUDGET (UNPACK(PRRS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),8,'ACC_BU_RRR')
  END IF
  IF (LBUDGET_RS) THEN
!$acc update self(PRSS)
    CALL BUDGET (UNPACK(PRSS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),10,'ACC_BU_RRS')
  END IF
  IF (LBUDGET_RG) THEN
!$acc update self(PRGS)
    CALL BUDGET (UNPACK(PRGS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),11,'ACC_BU_RRG')
  END IF
!
!*       5.3    Conversion-Melting of the aggregates
!
!$acc kernels
  GWORK(:) = PRST(:)>XRTMIN(5) .AND. PRSS(:)>0.0 .AND. PZT(:)>XTT
  WHERE( GWORK(:) )
    ZZW(:) = PRVT(:)*PPRES(:)/((XMV/XMD)+PRVT(:)) ! Vapor pressure
    ZZW(:) =  PKA(:)*(XTT-PZT(:)) +                                 &
               ( PDV(:)*(XLVTT + ( XCPV - XCL ) * ( PZT(:) - XTT )) &
                           *(XESTT-ZZW(:))/(XRV*PZT(:))             )
!
! compute RSMLT
!
#ifndef MNH_BITREP
    ZZW(:)  = MIN( PRSS(:), XFSCVMG*MAX( 0.0,( -ZZW(:) *             &
                           ( X0DEPS*       PLBDAS(:)**XEX0DEPS +     &
                             X1DEPS*PCJ(:)*PLBDAS(:)**XEX1DEPS ) ) / &
                                             ( PRHODREF(:)*XLMTT ) ) )
#else
    ZZW(:)  = MIN( PRSS(:), XFSCVMG*MAX( 0.0,( -ZZW(:) *                    &
                           ( X0DEPS*       BR_POW(PLBDAS(:),XEX0DEPS) +     &
                             X1DEPS*PCJ(:)*BR_POW(PLBDAS(:),XEX1DEPS) ) ) / &
                                             ( PRHODREF(:)*XLMTT ) ) )
#endif
!
! note that RSCVMG = RSMLT*XFSCVMG but no heat is exchanged (at the rate RSMLT)
! because the graupeln produced by this process are still icy!!!
!
    PRSS(:) = PRSS(:) - ZZW(:)
    PRGS(:) = PRGS(:) + ZZW(:)
  END WHERE
!$acc end kernels
  IF (LBUDGET_RS) THEN
!$acc update self(PRSS)
    CALL BUDGET (UNPACK(PRSS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),10,'CMEL_BU_RRS')
  END IF
  IF (LBUDGET_RG) THEN
!$acc update self(PRGS)
    CALL BUDGET (UNPACK(PRGS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),11,'CMEL_BU_RRG')
  END IF
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRCS,"RAIN_ICE_FAST_RS end:PRCS")
  CALL MPPDB_CHECK(PRRS,"RAIN_ICE_FAST_RS end:PRRS")
  CALL MPPDB_CHECK(PRSS,"RAIN_ICE_FAST_RS end:PRSS")
  CALL MPPDB_CHECK(PRGS,"RAIN_ICE_FAST_RS end:PRGS")
  CALL MPPDB_CHECK(PTHS,"RAIN_ICE_FAST_RS end:PTHS")
END IF

END SUBROUTINE RAIN_ICE_FAST_RS

END MODULE MODE_RAIN_ICE_FAST_RS
