!MNH_LIC Copyright 1995-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
! Modifications:
!  P. Wautelet 25/02/2019: split rain_ice (cleaner and easier to maintain/debug)
!-----------------------------------------------------------------
MODULE MODE_RAIN_ICE_SLOW

  IMPLICIT NONE

  PRIVATE

  PUBLIC :: RAIN_ICE_SLOW

CONTAINS

SUBROUTINE RAIN_ICE_SLOW(OMICRO, PINVTSTEP, PRHODREF,                      &
                         PRCT, PRRT, PRIT, PRST, PRGT, PRHODJ, PZT, PPRES, &
                         PLSFACT, PLVFACT,                                 &
                         PSSI, PRHODJ3D, PTHS3D, PRVS3D,                   &
                         PRVS, PRCS, PRRS, PRIS, PRSS, PRGS, PTHS,         &
                         PAI, PCJ, PKA, PDV, PLBDAS, PLBDAG)
!
!*      0. DECLARATIONS
!          ------------
!
use MODD_BUDGET,         only: LBUDGET_RC, LBUDGET_RG, LBUDGET_RI, LBUDGET_RR, LBUDGET_RS, LBUDGET_RV, LBUDGET_TH
use MODD_CST,            only: XALPI, XBETAI, XCI, XCPV, XGAMI, XLSTT, XMNH_HUGE_12_LOG, XP00, XRV, XTT
use MODD_RAIN_ICE_DESCR, only: XCEXVT, XLBDAS_MAX, XLBEXG, XLBEXS, XLBG, XLBS, XRTMIN
use MODD_RAIN_ICE_PARAM, only: X0DEPG, X0DEPS, X1DEPG, X1DEPS, XACRIAUTI, XALPHA3, XBCRIAUTI, XBETA3, XCOLEXIS, XCRIAUTI, &
                               XEX0DEPG, XEX0DEPS, XEX1DEPG, XEX1DEPS, XEXIAGGS, XFIAGGS, XHON, XSCFAC, XTEXAUTI, XTIMAUTI
!
use mode_mppdb
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
use MODI_BUDGET
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,  DIMENSION(:,:,:), intent(in)    :: OMICRO   ! Test where to compute all processes
REAL,                       intent(in)    :: PINVTSTEP
REAL,     DIMENSION(:),     intent(in)    :: PRHODREF ! RHO Dry REFerence
REAL,     DIMENSION(:),     intent(in)    :: PRCT     ! Cloud water m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRRT     ! Rain water m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRIT     ! Pristine ice m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRST     ! Snow/aggregate m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRGT     ! Graupel m.r. at t
REAL,     DIMENSION(:),     intent(in)    :: PRHODJ   ! RHO times Jacobian
REAL,     DIMENSION(:),     intent(in)    :: PZT      ! Temperature
REAL,     DIMENSION(:),     intent(in)    :: PPRES    ! Pressure
REAL,     DIMENSION(:),     intent(in)    :: PLSFACT  ! L_s/(Pi_ref*C_ph)
REAL,     DIMENSION(:),     intent(in)    :: PLVFACT  ! L_v/(Pi_ref*C_ph)
REAL,     DIMENSION(:),     intent(in)    :: PSSI     ! Supersaturation over ice
REAL,     DIMENSION(:,:,:), INTENT(IN)    :: PRHODJ3D ! Dry density * Jacobian
REAL,     DIMENSION(:,:,:), INTENT(IN)    :: PTHS3D   ! Theta source
REAL,     DIMENSION(:,:,:), INTENT(IN)    :: PRVS3D   ! Water vapor m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRVS     ! Water vapor m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRCS     ! Cloud water m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRRS     ! Rain water m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRIS     ! Pristine ice m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRSS     ! Snow/aggregate m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PRGS     ! Graupel m.r. source
REAL,     DIMENSION(:),     INTENT(INOUT) :: PTHS     ! Theta source
REAL,     DIMENSION(:),     intent(OUT)   :: PAI      ! Thermodynamical function
REAL,     DIMENSION(:),     intent(OUT)   :: PCJ      ! Function to compute the ventilation coefficient
REAL,     DIMENSION(:),     intent(OUT)   :: PKA      ! Thermal conductivity of the air
REAL,     DIMENSION(:),     intent(OUT)   :: PDV      ! Diffusivity of water vapor in the air
REAL,     DIMENSION(:),     intent(OUT)   :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL,     DIMENSION(:),     intent(OUT)   :: PLBDAG   ! Slope parameter of the graupel   distribution
!
! IN variables
!
!$acc declare present(OMICRO, PRHODREF, PRCT, PRRT, PRIT, PRST, PRGT, &
!$acc &               PRHODJ, PZT, PPRES, PLSFACT, PLVFACT, PSSI,     &
!$acc &               PRHODJ3D, PTHS3D, PRVS3D)                       &
!
! INOUT variables
!
!$acc &       present(PRVS, PRCS, PRRS, PRIS, PRSS, PRGS, PTHS)       &
!
! OUT variables
!
!$acc &       present(PAI, PCJ, PKA, PDV, PLBDAS, PLBDAG)
!
!*       0.2  declaration of local variables
!
LOGICAL, DIMENSION(:), ALLOCATABLE :: GWORK
REAL,    DIMENSION(:), ALLOCATABLE :: ZCRIAUTI ! Snow-to-ice autoconversion thres.
REAL,    DIMENSION(:), ALLOCATABLE :: ZZW      ! Work array
!
!$acc declare device_resident(GWORK, ZZW, ZCRIAUTI)
!
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(OMICRO,"RAIN_ICE_SLOW beg:OMICRO")
  CALL MPPDB_CHECK(PRHODREF,"RAIN_ICE_SLOW beg:PRHODREF")
  CALL MPPDB_CHECK(PRCT,"RAIN_ICE_SLOW beg:PRCT")
  CALL MPPDB_CHECK(PRRT,"RAIN_ICE_SLOW beg:PRRT")
  CALL MPPDB_CHECK(PRIT,"RAIN_ICE_SLOW beg:PRIT")
  CALL MPPDB_CHECK(PRST,"RAIN_ICE_SLOW beg:PRST")
  CALL MPPDB_CHECK(PRGT,"RAIN_ICE_SLOW beg:PRGT")
  CALL MPPDB_CHECK(PRHODJ,"RAIN_ICE_SLOW beg:PRHODJ")
  CALL MPPDB_CHECK(PZT,"RAIN_ICE_SLOW beg:PZT")
  CALL MPPDB_CHECK(PPRES,"RAIN_ICE_SLOW beg:PPRES")
  CALL MPPDB_CHECK(PLSFACT,"RAIN_ICE_SLOW beg:PLSFACT")
  CALL MPPDB_CHECK(PLVFACT,"RAIN_ICE_SLOW beg:PLVFACT")
  CALL MPPDB_CHECK(PSSI,"RAIN_ICE_SLOW beg:PSSI")
  CALL MPPDB_CHECK(PRHODJ3D,"RAIN_ICE_SLOW beg:PRHODJ3D")
  CALL MPPDB_CHECK(PTHS3D,"RAIN_ICE_SLOW beg:PTHS3D")
  CALL MPPDB_CHECK(PRVS3D,"RAIN_ICE_SLOW beg:PRVS3D")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRVS,"RAIN_ICE_SLOW beg:PRVS")
  CALL MPPDB_CHECK(PRCS,"RAIN_ICE_SLOW beg:PRCS")
  CALL MPPDB_CHECK(PRIS,"RAIN_ICE_SLOW beg:PRIS")
  CALL MPPDB_CHECK(PRSS,"RAIN_ICE_SLOW beg:PRSS")
  CALL MPPDB_CHECK(PRGS,"RAIN_ICE_SLOW beg:PRGS")
  CALL MPPDB_CHECK(PTHS,"RAIN_ICE_SLOW beg:PTHS")
END IF
!
ALLOCATE( GWORK   (size(PRHODREF)) )
ALLOCATE( ZZW     (size(PRHODREF)) )
ALLOCATE( ZCRIAUTI(size(PRHODREF)) )
!
!*       3.2     compute the homogeneous nucleation source: RCHONI
!
!$acc kernels
  ZZW(:) = 0.0
  GWORK(:) = PZT(:)<XTT-35.0 .AND. PRCT(:)>XRTMIN(2) .AND. PRCS(:)>0.
  WHERE( GWORK(:) )
    ZZW(:) = MIN( PRCS(:),XHON*PRHODREF(:)*PRCT(:)       &
#ifndef MNH_BITREP
                                 *EXP( MIN(XMNH_HUGE_12_LOG,XALPHA3*(PZT(:)-XTT)-XBETA3) ) )
#else
                                 *BR_EXP(MIN(XMNH_HUGE_12_LOG, XALPHA3*(PZT(:)-XTT)-XBETA3) ) )
#endif
    PRIS(:) = PRIS(:) + ZZW(:)
    PRCS(:) = PRCS(:) - ZZW(:)
    PTHS(:) = PTHS(:) + ZZW(:)*(PLSFACT(:)-PLVFACT(:)) ! f(L_f*(RCHONI))
  ENDWHERE
!$acc end kernels
!
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'HON_BU_RTH')
  END IF
  IF (LBUDGET_RC) THEN
!$acc update self(PRCS)
    CALL BUDGET (UNPACK(PRCS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),7,'HON_BU_RRC')
  END IF
  IF (LBUDGET_RI) THEN
!$acc update self(PRIS)
    CALL BUDGET (UNPACK(PRIS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),9,'HON_BU_RRI')
  END IF
!
!*       3.3     compute the spontaneous freezing source: RRHONG
!
!$acc kernels
  ZZW(:) = 0.0
  GWORK(:) = PZT(:)<XTT-35.0 .AND. PRRT(:)>XRTMIN(3) .AND. PRRS(:)>0.
  WHERE( GWORK(:) )
    ZZW(:) = MIN( PRRS(:),PRRT(:)* PINVTSTEP )
    PRGS(:) = PRGS(:) + ZZW(:)
    PRRS(:) = PRRS(:) - ZZW(:)
    PTHS(:) = PTHS(:) + ZZW(:)*(PLSFACT(:)-PLVFACT(:)) ! f(L_f*(RRHONG))
  ENDWHERE
!$acc end kernels
!
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'SFR_BU_RTH')
  END IF
  IF (LBUDGET_RR) THEN
!$acc update self(PRRS)
    CALL BUDGET (UNPACK(PRRS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),8,'SFR_BU_RRR')
  END IF
  IF (LBUDGET_RG) THEN
!$acc update self(PRGS)
    CALL BUDGET (UNPACK(PRGS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),11,'SFR_BU_RRG')
  END IF
!
!*       3.4    compute the deposition, aggregation and autoconversion sources
!
!$acc kernels
  PKA(:) = 2.38E-2 + 0.0071E-2 * ( PZT(:) - XTT )          ! k_a
#ifndef MNH_BITREP
  PDV(:) = 0.211E-4 * (PZT(:)/XTT)**1.94 * (XP00/PPRES(:)) ! D_v
#else
  PDV(:) = 0.211E-4 * BR_POW(PZT(:)/XTT,1.94) * (XP00/PPRES(:)) ! D_v
#endif
!
!*       3.4.1  compute the thermodynamical function A_i(T,P)
!*              and the c^prime_j (in the ventilation factor)
!

#ifndef MNH_BITREP
  PAI(:) = EXP( XALPI - XBETAI/PZT(:) - XGAMI*ALOG(PZT(:) ) ) ! es_i
  PAI(:) = ( XLSTT + (XCPV-XCI)*(PZT(:)-XTT) )**2 / (PKA(:)*XRV*PZT(:)**2) &
                                 + ( XRV*PZT(:) ) / (PDV(:)*PAI(:))
  PCJ(:) = XSCFAC * PRHODREF(:)**0.3 / SQRT( 1.718E-5+0.0049E-5*(PZT(:)-XTT) )
#else
  PAI(:) = BR_EXP( XALPI - XBETAI/PZT(:) - XGAMI*BR_LOG(PZT(:) ) ) ! es_i
  PAI(:) = BR_P2( XLSTT + (XCPV-XCI)*(PZT(:)-XTT) ) / (PKA(:)*XRV*BR_P2(PZT(:))) &
                                 + ( XRV*PZT(:) ) / (PDV(:)*PAI(:))
  PCJ(:) = XSCFAC * BR_POW(PRHODREF(:),0.3) / BR_POW( 1.718E-5+0.0049E-5*(PZT(:)-XTT) , 0.5)
#endif
!
!*       3.4.2  compute the riming-conversion of r_c for r_i production: RCAUTI
!
!  ZZW(:) = 0.0
!  ZTIMAUTIC = SQRT( XTIMAUTI*XTIMAUTC )
!  WHERE ( (PRCT(:)>0.0) .AND. (PRIT(:)>0.0) .AND. (PRCS(:)>0.0) )
!    ZZW(:) = MIN( PRCS(:),ZTIMAUTIC * MAX( SQRT( PRIT(:)*PRCT(:) ),0.0 ) )
!    PRIS(:) = PRIS(:) + ZZW(:)
!    PRCS(:) = PRCS(:) - ZZW(:)
!    PTHS(:) = PTHS(:) + ZZW(:)*(PLSFACT(:)-PLVFACT(:)) ! f(L_f*(RCAUTI))
!  END WHERE
!
!*       3.4.3  compute the deposition on r_s: RVDEPS
!
  GWORK(:) = PRST(:)>0.0
  WHERE ( GWORK(:) )
#ifndef MNH_BITREP
    PLBDAS(:)  = MIN( XLBDAS_MAX,                                           &
                      XLBS*( PRHODREF(:)*MAX( PRST(:),XRTMIN(5) ) )**XLBEXS )
#else
    PLBDAS(:)  = MIN( XLBDAS_MAX,                                           &
                      XLBS*BR_POW( PRHODREF(:)*MAX( PRST(:),XRTMIN(5) ),XLBEXS ) )
#endif
  ELSEWHERE
    PLBDAS(:) = 0.
  END WHERE
  ZZW(:) = 0.0
  GWORK(:) = (PRST(:)>XRTMIN(5)) .AND. (PRSS(:)>0.0)
  WHERE ( GWORK(:) )
#ifndef MNH_BITREP
    ZZW(:) = ( PSSI(:)/(PRHODREF(:)*PAI(:)) ) *                               &
             ( X0DEPS*PLBDAS(:)**XEX0DEPS + X1DEPS*PCJ(:)*PLBDAS(:)**XEX1DEPS )
#else
    ZZW(:) = ( PSSI(:)/(PRHODREF(:)*PAI(:)) ) *                               &
             ( X0DEPS*BR_POW(PLBDAS(:),XEX0DEPS) + X1DEPS*PCJ(:)*BR_POW(PLBDAS(:),XEX1DEPS) )
#endif
    ZZW(:) =         MIN( PRVS(:),ZZW(:)      )*(0.5+SIGN(0.5,ZZW(:))) &
                   - MIN( PRSS(:),ABS(ZZW(:)) )*(0.5-SIGN(0.5,ZZW(:)))
    PRSS(:) = PRSS(:) + ZZW(:)
    PRVS(:) = PRVS(:) - ZZW(:)
    PTHS(:) = PTHS(:) + ZZW(:)*PLSFACT(:)
  END WHERE
!$acc end kernels
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'DEPS_BU_RTH')
  END IF
  IF (LBUDGET_RV) THEN
!$acc update self(PRVS)
    CALL BUDGET (UNPACK(PRVS(:),MASK=OMICRO(:,:,:),FIELD=PRVS3D)*PRHODJ3D(:,:,:),6,'DEPS_BU_RRV')
  END IF
  IF (LBUDGET_RS) THEN
!$acc update self(PRSS)
    CALL BUDGET (UNPACK(PRSS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),10,'DEPS_BU_RRS')
  END IF
!
!*       3.4.4  compute the aggregation on r_s: RIAGGS
!
!$acc kernels
  ZZW(:) = 0.0
  GWORK(:) = PRIT(:)>XRTMIN(4) .AND. PRST(:)>XRTMIN(5) .AND. PRIS(:)>0.0
  WHERE ( GWORK(:) )
#ifndef MNH_BITREP
    ZZW(:) = MIN( PRIS(:),XFIAGGS * EXP( XCOLEXIS*(PZT(:)-XTT) ) &
                                  * PRIT(:)                      &
                                  * PLBDAS(:)**XEXIAGGS          &
                                  * PRHODREF(:)**(-XCEXVT)       )
#else
    ZZW(:) = MIN( PRIS(:),XFIAGGS * BR_EXP( XCOLEXIS*(PZT(:)-XTT) ) &
                                  * PRIT(:)                         &
                                  * BR_POW(PLBDAS(:),XEXIAGGS)      &
                                  * BR_POW(PRHODREF(:),-XCEXVT)     )
#endif
    PRSS(:)  = PRSS(:)  + ZZW(:)
    PRIS(:)  = PRIS(:)  - ZZW(:)
  END WHERE
!$acc end kernels
  IF (LBUDGET_RI) THEN
!$acc update self(PRIS)
    CALL BUDGET (UNPACK(PRIS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),9,'AGGS_BU_RRI')
  END IF
  IF (LBUDGET_RS) THEN
!$acc update self(PRSS)
    CALL BUDGET (UNPACK(PRSS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),10,'AGGS_BU_RRS')
  END IF
!
!*       3.4.5  compute the autoconversion of r_i for r_s production: RIAUTS
!
!$acc kernels
#ifndef MNH_BITREP
  ZCRIAUTI(:)=MIN(XCRIAUTI,10**(XACRIAUTI*(PZT(:)-XTT)+XBCRIAUTI))
#else
  ZCRIAUTI(:)=MIN(XCRIAUTI, BR_POW(10.,XACRIAUTI*(PZT(:)-XTT)+XBCRIAUTI) )
#endif
  ZZW(:) = 0.0
  GWORK(:) = PRIT(:)>XRTMIN(4) .AND. PRIS(:)>0.0
  WHERE ( GWORK(:) )
#ifndef MNH_BITREP
    ZZW(:) = MIN( PRIS(:),XTIMAUTI * EXP( XTEXAUTI*(PZT(:)-XTT) ) &
                            * MAX( PRIT(:)-ZCRIAUTI(:),0.0 ) )
#else
    ZZW(:) = MIN( PRIS(:),XTIMAUTI * BR_EXP( XTEXAUTI*(PZT(:)-XTT) ) &
                            * MAX( PRIT(:)-ZCRIAUTI(:),0.0 ) )
#endif
    PRSS(:)  = PRSS(:)  + ZZW(:)
    PRIS(:)  = PRIS(:)  - ZZW(:)
  END WHERE
!$acc end kernels
  IF (LBUDGET_RI) THEN
!$acc update self(PRIS)
    CALL BUDGET (UNPACK(PRIS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),9,'AUTS_BU_RRI')
  END IF
  IF (LBUDGET_RS) THEN
!$acc update self(PRSS)
    CALL BUDGET (UNPACK(PRSS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),10,'AUTS_BU_RRS')
  END IF
!
!*       3.4.6  compute the deposition on r_g: RVDEPG
!
!
!$acc kernels
  GWORK(:) = PRGT(:)>0.0
  WHERE ( GWORK(:) )
#ifndef MNH_BITREP
    PLBDAG(:)  = XLBG*( PRHODREF(:)*MAX( PRGT(:),XRTMIN(6) ) )**XLBEXG
#else
    PLBDAG(:)  = XLBG*BR_POW( PRHODREF(:)*MAX( PRGT(:),XRTMIN(6) ), XLBEXG)
#endif
  ELSEWHERE
    PLBDAG(:) = 0.
  END WHERE
  ZZW(:) = 0.0
  GWORK(:) = PRGT(:)>XRTMIN(6) .AND. PRGS(:)>0.0
  WHERE ( GWORK(:) )
#ifndef MNH_BITREP
    ZZW(:) = ( PSSI(:)/(PRHODREF(:)*PAI(:)) ) *                               &
             ( X0DEPG*PLBDAG(:)**XEX0DEPG + X1DEPG*PCJ(:)*PLBDAG(:)**XEX1DEPG )
#else
    ZZW(:) = ( PSSI(:)/(PRHODREF(:)*PAI(:)) ) *                               &
             ( X0DEPG*BR_POW(PLBDAG(:),XEX0DEPG) + X1DEPG*PCJ(:)*BR_POW(PLBDAG(:),XEX1DEPG) )
#endif
    ZZW(:) =         MIN( PRVS(:),ZZW(:)      )*(0.5+SIGN(0.5,ZZW(:))) &
                   - MIN( PRGS(:),ABS(ZZW(:)) )*(0.5-SIGN(0.5,ZZW(:)))
    PRGS(:) = PRGS(:) + ZZW(:)
    PRVS(:) = PRVS(:) - ZZW(:)
    PTHS(:) = PTHS(:) + ZZW(:)*PLSFACT(:)
  END WHERE
!$acc end kernels
  IF (LBUDGET_TH) THEN
!$acc update self(PTHS)
    CALL BUDGET (UNPACK(PTHS(:),MASK=OMICRO(:,:,:),FIELD=PTHS3D)*PRHODJ3D(:,:,:),4,'DEPG_BU_RTH')
  END IF
  IF (LBUDGET_RV) THEN
!$acc update self(PRVS)
    CALL BUDGET (UNPACK(PRVS(:),MASK=OMICRO(:,:,:),FIELD=PRVS3D)*PRHODJ3D(:,:,:),6,'DEPG_BU_RRV')
  END IF
  IF (LBUDGET_RG) THEN
!$acc update self(PRGS)
    CALL BUDGET (UNPACK(PRGS(:)*PRHODJ(:),MASK=OMICRO(:,:,:),FIELD=0.0),11,'DEPG_BU_RRG')
  END IF
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRVS,"RAIN_ICE_SLOW end:PRVS")
  CALL MPPDB_CHECK(PRCS,"RAIN_ICE_SLOW end:PRCS")
  CALL MPPDB_CHECK(PRIS,"RAIN_ICE_SLOW end:PRIS")
  CALL MPPDB_CHECK(PRSS,"RAIN_ICE_SLOW end:PRSS")
  CALL MPPDB_CHECK(PRGS,"RAIN_ICE_SLOW end:PRGS")
  CALL MPPDB_CHECK(PTHS,"RAIN_ICE_SLOW end:PTHS")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PAI,"RAIN_ICE_SLOW end:PAI")
  CALL MPPDB_CHECK(PCJ,"RAIN_ICE_SLOW end:PCJ")
  CALL MPPDB_CHECK(PKA,"RAIN_ICE_SLOW end:PKA")
  CALL MPPDB_CHECK(PDV,"RAIN_ICE_SLOW end:PDV")
  CALL MPPDB_CHECK(PLBDAS,"RAIN_ICE_SLOW end:PLBDAS")
  CALL MPPDB_CHECK(PLBDAG,"RAIN_ICE_SLOW end:PLBDAG")
END IF

END SUBROUTINE RAIN_ICE_SLOW

END MODULE MODE_RAIN_ICE_SLOW
